__all__ = [
    'Tweet', 'Medium', 'Photo', 'VideoVariant', 'Video', 'Gif', 'TextLink', 'Coordinates', 'Place',
    'User', 'UserLabel',
    'Trend',
    'GuestTokenManager',
    'TwitterSearchScraper',
    'TwitterUserScraper',
    'TwitterProfileScraper',
    'TwitterHashtagScraper',
    'TwitterTweetScraperMode',
    'TwitterTweetScraper',
    'TwitterListPostsScraper',
    'TwitterTrendsScraper',
]

import collections
import copy
import dataclasses
import datetime
import email.utils
import enum
import filelock
import itertools
import json
import random
import logging
import os
import re
import requests.adapters
import umbraSnscrpe.base
import string
import time
import typing
import urllib.parse
import urllib3.util.ssl_
import warnings

# DescriptionURL deprecation
_DEPRECATED_NAMES = {'DescriptionURL': 'TextLink'}


def __getattr__(name):
    if name in _DEPRECATED_NAMES:
        warnings.warn(f'{name} is deprecated, use {_DEPRECATED_NAMES[name]} instead', FutureWarning, stacklevel=2)
        return globals()[_DEPRECATED_NAMES[name]]
    raise AttributeError(f'module {__name__!r} has no attribute {name!r}')


def __dir__():
    return sorted(__all__ + list(_DEPRECATED_NAMES.keys()))


_logger = logging.getLogger(__name__)
_API_AUTHORIZATION_HEADER = 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs=1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA'
_globalGuestTokenManager = None
_GUEST_TOKEN_VALIDITY = 10800
_CIPHERS_CHROME = 'TLS_AES_128_GCM_SHA256:TLS_AES_256_GCM_SHA384:TLS_CHACHA20_POLY1305_SHA256:ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305:ECDHE-RSA-AES128-SHA:ECDHE-RSA-AES256-SHA:AES128-GCM-SHA256:AES256-GCM-SHA384:AES128-SHA:AES256-SHA'


@dataclasses.dataclass
class Tweet(umbraSnscrpe.base.Item):
    url: str
    date: datetime.datetime
    rawContent: str
    renderedContent: str
    id: int
    user: 'User'
    replyCount: int
    retweetCount: int
    likeCount: int
    quoteCount: int
    conversationId: int
    lang: str
    source: typing.Optional[str] = None
    sourceUrl: typing.Optional[str] = None
    sourceLabel: typing.Optional[str] = None
    links: typing.Optional[typing.List['TextLink']] = None
    media: typing.Optional[typing.List['Medium']] = None
    retweetedTweet: typing.Optional['Tweet'] = None
    quotedTweet: typing.Optional['Tweet'] = None
    inReplyToTweetId: typing.Optional[int] = None
    inReplyToUser: typing.Optional['User'] = None
    mentionedUsers: typing.Optional[typing.List['User']] = None
    coordinates: typing.Optional['Coordinates'] = None
    place: typing.Optional['Place'] = None
    hashtags: typing.Optional[typing.List[str]] = None
    cashtags: typing.Optional[typing.List[str]] = None
    card: typing.Optional['Card'] = None
    viewCount: typing.Optional[int] = None
    vibe: typing.Optional['Vibe'] = None

    username = umbraSnscrpe.base._DeprecatedProperty('username', lambda self: self.user.username, 'user.username')
    outlinks = umbraSnscrpe.base._DeprecatedProperty('outlinks',
                                                     lambda self: [x.url for x in self.links] if self.links else [],
                                                 'links (url attribute)')
    outlinksss = umbraSnscrpe.base._DeprecatedProperty('outlinksss', lambda self: ' '.join(
        x.url for x in self.links) if self.links else '', 'links (url attribute)')
    tcooutlinks = umbraSnscrpe.base._DeprecatedProperty('tcooutlinks',
                                                        lambda self: [x.tcourl for x in self.links] if self.links else [],
                                                    'links (tcourl attribute)')
    tcooutlinksss = umbraSnscrpe.base._DeprecatedProperty('tcooutlinksss', lambda self: ' '.join(
        x.tcourl for x in self.links) if self.links else '', 'links (tcourl attribute)')
    content = umbraSnscrpe.base._DeprecatedProperty('content', lambda self: self.rawContent, 'rawContent')

    def __str__(self):
        return self.url


@dataclasses.dataclass
class TextLink:
    text: typing.Optional[str]
    url: str
    tcourl: str
    indices: typing.Tuple[int, int]


class Medium:
    pass


@dataclasses.dataclass
class Photo(Medium):
    previewUrl: str
    fullUrl: str
    altText: typing.Optional[str] = None


@dataclasses.dataclass
class VideoVariant:
    url: str
    contentType: typing.Optional[str]
    bitrate: typing.Optional[int]


@dataclasses.dataclass
class Video(Medium):
    thumbnailUrl: str
    variants: typing.List[VideoVariant]
    duration: typing.Optional[float] = None
    views: typing.Optional[int] = None
    altText: typing.Optional[str] = None


@dataclasses.dataclass
class Gif(Medium):
    thumbnailUrl: str
    variants: typing.List[VideoVariant]
    altText: typing.Optional[str] = None


@dataclasses.dataclass
class Coordinates:
    longitude: float
    latitude: float


@dataclasses.dataclass
class Place:
    id: str
    fullName: str
    name: str
    type: str
    country: str
    countryCode: str


class Card:
    pass


@dataclasses.dataclass
class SummaryCard(Card):
    title: str
    url: str
    description: typing.Optional[str] = None
    thumbnailUrl: typing.Optional[str] = None
    siteUser: typing.Optional['User'] = None
    creatorUser: typing.Optional['User'] = None


@dataclasses.dataclass
class AppCard(SummaryCard):
    pass


@dataclasses.dataclass
class PollCard(Card):
    options: typing.List['PollOption']
    endDate: datetime.datetime
    duration: int
    finalResults: bool
    lastUpdateDate: typing.Optional[datetime.datetime] = None
    medium: typing.Optional[Medium] = None


@dataclasses.dataclass
class PollOption:
    label: str
    count: typing.Optional[int] = None


@dataclasses.dataclass
class PlayerCard(Card):
    title: str
    url: str
    description: typing.Optional[str] = None
    imageUrl: typing.Optional[str] = None
    siteUser: typing.Optional['User'] = None


@dataclasses.dataclass
class PromoConvoCard(Card):
    actions: typing.List['PromoConvoAction']
    thankYouText: str
    medium: Medium
    thankYouUrl: typing.Optional[str] = None
    thankYouTcoUrl: typing.Optional[str] = None
    cover: typing.Optional['Photo'] = None


@dataclasses.dataclass
class PromoConvoAction:
    label: str
    tweet: str


@dataclasses.dataclass
class BroadcastCard(Card):
    id: str
    url: str
    title: str
    state: typing.Optional[str] = None
    broadcaster: typing.Optional['User'] = None
    thumbnailUrl: typing.Optional[str] = None
    source: typing.Optional[str] = None
    siteUser: typing.Optional['User'] = None


@dataclasses.dataclass
class PeriscopeBroadcastCard(Card):
    id: str
    url: str
    title: str
    description: str
    state: str
    totalParticipants: int
    thumbnailUrl: typing.Optional[str] = None
    source: typing.Optional[str] = None
    broadcaster: typing.Optional['User'] = None
    siteUser: typing.Optional['User'] = None


@dataclasses.dataclass
class EventCard(Card):
    event: 'Event'


@dataclasses.dataclass
class Event:
    id: int
    category: str
    photo: Photo
    title: typing.Optional[str] = None
    description: typing.Optional[str] = None

    @property
    def url(self):
        return f'https://twitter.com/i/events/{self.id}'


@dataclasses.dataclass
class NewsletterCard(Card):
    title: str
    description: str
    url: str
    revueAccountId: int
    issueCount: int
    imageUrl: typing.Optional[str] = None


@dataclasses.dataclass
class NewsletterIssueCard(Card):
    newsletterTitle: str
    newsletterDescription: str
    issueTitle: str
    issueNumber: int
    url: str
    revueAccountId: int
    issueDescription: typing.Optional[str] = None
    imageUrl: typing.Optional[str] = None


@dataclasses.dataclass
class AmplifyCard(Card):
    id: str
    video: Video


@dataclasses.dataclass
class AppPlayerCard(Card):
    title: str
    video: Video
    appCategory: str
    playerOwnerId: int
    siteUser: typing.Optional['User'] = None


@dataclasses.dataclass
class SpacesCard(Card):
    url: str
    id: str


@dataclasses.dataclass
class MessageMeCard(Card):
    recipient: 'User'
    url: str
    buttonText: str


UnifiedCardComponentKey = str
UnifiedCardDestinationKey = str
UnifiedCardMediumKey = str
UnifiedCardAppKey = str


@dataclasses.dataclass
class UnifiedCard(Card):
    componentObjects: typing.Dict[UnifiedCardComponentKey, 'UnifiedCardComponentObject']
    destinations: typing.Dict[UnifiedCardDestinationKey, 'UnifiedCardDestination']
    media: typing.Dict[UnifiedCardMediumKey, Medium]
    apps: typing.Optional[typing.Dict[UnifiedCardAppKey, typing.List['UnifiedCardApp']]] = None
    components: typing.Optional[typing.List[UnifiedCardComponentKey]] = None
    swipeableLayoutSlides: typing.Optional[typing.List['UnifiedCardSwipeableLayoutSlide']] = None
    collectionLayoutSlides: typing.Optional[typing.List['UnifiedCardCollectionLayoutSlide']] = None
    type: typing.Optional[str] = None

    def __post_init__(self):
        if (self.components is not None) + (self.swipeableLayoutSlides is not None) + (
                self.collectionLayoutSlides is not None) != 1:
            raise ValueError('did not get exactly one of components, swipeableLayoutSlides, and collectionLayoutSlides')
        if self.components and not all(k in self.componentObjects for k in self.components):
            raise ValueError('missing components')
        if self.swipeableLayoutSlides and not all(
                s.mediumComponentKey in self.componentObjects and s.componentKey in self.componentObjects for s in
                self.swipeableLayoutSlides):
            raise ValueError('missing components')
        if any(c.destinationKey not in self.destinations for c in self.componentObjects.values() if
               hasattr(c, 'destinationKey')):
            raise ValueError('missing destinations')
        if any(b.destinationKey not in self.destinations for c in self.componentObjects.values() if
               isinstance(c, UnifiedCardButtonGroupComponentObject) for b in c.buttons):
            raise ValueError('missing destinations')
        mediaKeys = []
        for c in self.componentObjects.values():
            if isinstance(c, UnifiedCardMediumComponentObject):
                mediaKeys.append(c.mediumKey)
            elif isinstance(c, UnifiedCardSwipeableMediaComponentObject):
                mediaKeys.extend(x.mediumKey for x in c.media)
        mediaKeys.extend(d.mediumKey for d in self.destinations.values() if d.mediumKey is not None)
        mediaKeys.extend(a.iconMediumKey for l in (self.apps.values() if self.apps is not None else []) for a in l if
                         a.iconMediumKey is not None)
        if any(k not in self.media for k in mediaKeys):
            raise ValueError('missing media')
        if any(c.appKey not in self.apps for c in self.componentObjects.values() if hasattr(c, 'appKey')):
            raise ValueError('missing apps')
        if any(d.appKey not in self.apps for d in self.destinations.values() if d.appKey is not None):
            raise ValueError('missing apps')


class UnifiedCardComponentObject:
    pass


@dataclasses.dataclass
class UnifiedCardDetailComponentObject(UnifiedCardComponentObject):
    content: str
    destinationKey: UnifiedCardDestinationKey


@dataclasses.dataclass
class UnifiedCardMediumComponentObject(UnifiedCardComponentObject):
    mediumKey: UnifiedCardMediumKey
    destinationKey: UnifiedCardDestinationKey


@dataclasses.dataclass
class UnifiedCardButtonGroupComponentObject(UnifiedCardComponentObject):
    buttons: typing.List['UnifiedCardButton']


@dataclasses.dataclass
class UnifiedCardButton:
    text: str
    destinationKey: UnifiedCardDestinationKey


@dataclasses.dataclass
class UnifiedCardSwipeableMediaComponentObject(UnifiedCardComponentObject):
    media: typing.List['UnifiedCardSwipeableMediaMedium']


@dataclasses.dataclass
class UnifiedCardSwipeableMediaMedium:
    mediumKey: UnifiedCardMediumKey
    destinationKey: UnifiedCardDestinationKey


@dataclasses.dataclass
class UnifiedCardAppStoreComponentObject(UnifiedCardComponentObject):
    appKey: UnifiedCardAppKey
    destinationKey: UnifiedCardDestinationKey


@dataclasses.dataclass
class UnifiedCardTwitterListDetailsComponentObject(UnifiedCardComponentObject):
    name: str
    memberCount: int
    subscriberCount: int
    user: 'User'
    destinationKey: UnifiedCardDestinationKey


@dataclasses.dataclass
class UnifiedCardTwitterCommunityDetailsComponentObject(UnifiedCardComponentObject):
    name: str
    theme: str
    membersCount: int
    destinationKey: UnifiedCardDestinationKey
    membersFacepile: typing.Optional[typing.List['User']] = None


@dataclasses.dataclass
class UnifiedCardDestination:
    url: typing.Optional[str] = None
    appKey: typing.Optional[UnifiedCardAppKey] = None
    mediumKey: typing.Optional[UnifiedCardMediumKey] = None

    def __post_init__(self):
        if (self.url is None) == (self.appKey is None):
            raise ValueError('did not get exactly one of url and appKey')


@dataclasses.dataclass
class UnifiedCardApp:
    type: str
    id: str
    title: str
    countryCode: str
    url: str
    category: typing.Optional[str] = None
    description: typing.Optional[str] = None
    iconMediumKey: typing.Optional[UnifiedCardMediumKey] = None
    size: typing.Optional[int] = None
    installs: typing.Optional[int] = None
    ratingAverage: typing.Optional[float] = None
    ratingCount: typing.Optional[int] = None
    isFree: typing.Optional[bool] = None
    isEditorsChoice: typing.Optional[bool] = None
    hasInAppPurchases: typing.Optional[bool] = None
    hasInAppAds: typing.Optional[bool] = None


@dataclasses.dataclass
class UnifiedCardSwipeableLayoutSlide:
    mediumComponentKey: UnifiedCardComponentKey
    componentKey: UnifiedCardComponentKey


@dataclasses.dataclass
class UnifiedCardCollectionLayoutSlide:
    detailsComponentKey: UnifiedCardComponentKey
    mediumComponentKey: UnifiedCardComponentKey


@dataclasses.dataclass
class Vibe:
    text: str
    imageUrl: str
    imageDescription: str


@dataclasses.dataclass
class TweetRef(umbraSnscrpe.base.Item):
    '''A reference to a tweet for which no proper Tweet object could be produced from the data returned by Twitter'''

    id: int

    def __str__(self):
        return f'https://twitter.com/i/web/status/{self.id}'


@dataclasses.dataclass
class User(umbraSnscrpe.base.Entity):
    # Most fields can be None if they're not known.

    username: str
    id: int
    displayname: typing.Optional[str] = None
    rawDescription: typing.Optional[str] = None  # Raw description with the URL(s) intact
    renderedDescription: typing.Optional[
        str] = None  # Description as it's displayed on the web interface with URLs replaced
    descriptionLinks: typing.Optional[typing.List[TextLink]] = None
    verified: typing.Optional[bool] = None
    created: typing.Optional[datetime.datetime] = None
    followersCount: typing.Optional[int] = None
    friendsCount: typing.Optional[int] = None
    statusesCount: typing.Optional[int] = None
    favouritesCount: typing.Optional[int] = None
    listedCount: typing.Optional[int] = None
    mediaCount: typing.Optional[int] = None
    location: typing.Optional[str] = None
    protected: typing.Optional[bool] = None
    link: typing.Optional[TextLink] = None
    profileImageUrl: typing.Optional[str] = None
    profileBannerUrl: typing.Optional[str] = None
    label: typing.Optional['UserLabel'] = None

    descriptionUrls = umbraSnscrpe.base._DeprecatedProperty('descriptionUrls', lambda self: self.descriptionLinks,
                                                        'descriptionLinks')
    linkUrl = umbraSnscrpe.base._DeprecatedProperty('linkUrl', lambda self: self.link.url if self.link else None,
                                                'link.url')
    linkTcourl = umbraSnscrpe.base._DeprecatedProperty('linkTcourl', lambda self: self.link.tcourl if self.link else None,
                                                   'link.tcourl')
    description = umbraSnscrpe.base._DeprecatedProperty('description', lambda self: self.renderedDescription,
                                                    'renderedDescription')

    @property
    def url(self):
        return f'https://twitter.com/{self.username}'

    def __str__(self):
        return self.url


@dataclasses.dataclass
class UserLabel:
    description: str
    url: typing.Optional[str] = None
    badgeUrl: typing.Optional[str] = None
    longDescription: typing.Optional[str] = None


@dataclasses.dataclass
class UserRef:
    id: int


@dataclasses.dataclass
class Trend(umbraSnscrpe.base.Item):
    name: str
    domainContext: str
    metaDescription: typing.Optional[str] = None

    def __str__(self):
        return f'https://twitter.com/search?q={urllib.parse.quote(self.name)}'


class _ScrollDirection(enum.Enum):
    TOP = enum.auto()
    BOTTOM = enum.auto()
    BOTH = enum.auto()


class GuestTokenManager:
    def __init__(self):
        self._token = None
        self._setTime = 0.0

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, token):
        self._token = token
        self._setTime = time.time()

    @property
    def setTime(self):
        return self._setTime

    def reset(self):
        self._token = None
        self._setTime = 0.0


class _CLIGuestTokenManager(GuestTokenManager):
    def __init__(self):
        super().__init__()
        cacheHome = os.environ.get('XDG_CACHE_HOME')
        if not cacheHome or not os.path.isabs(cacheHome):
            # This should be ${HOME}/.cache, but the HOME environment variable may not exist on non-POSIX-compliant systems.
            # On POSIX-compliant systems, the XDG Base Directory specification is followed exactly since ~ expands to $HOME if it is present.
            cacheHome = os.path.join(os.path.expanduser('~'), '.cache')
        dir = os.path.join(cacheHome, 'umbraSnscrpe')
        if not os.path.isdir(dir):
            # os.makedirs does not apply mode recursively anymore. https://bugs.python.org/issue42367
            # This ensures that the XDG_CACHE_HOME is created with the right permissions.
            os.makedirs(os.path.dirname(dir), mode=0o700, exist_ok=True)
            os.mkdir(dir, mode=0o700)
        self._file = os.path.join(dir, 'cli-twitter-guest-token.json')
        self._lockFile = f'{self._file}.lock'
        self._lock = filelock.FileLock(self._lockFile)

    def _read(self):
        with self._lock:
            if not os.path.exists(self._file):
                return None
            _logger.info(f'Reading guest token from {self._file}')
            with open(self._file, 'r') as fp:
                try:
                    o = json.load(fp)
                except json.JSONDecodeError as e:
                    _logger.warning(f'Malformed guest token file {self._file}: {e!s}')
                    self.reset()
                    return None
        self._token = o['token']
        self._setTime = o['setTime']
        if self._setTime < time.time() - _GUEST_TOKEN_VALIDITY:
            _logger.info('Guest token expired')
            self.reset()

    def _write(self):
        with self._lock:
            _logger.info(f'Writing guest token to {self._file}')
            with open(self._file, 'w') as fp:
                json.dump({'token': self.token, 'setTime': self.setTime}, fp)

    @property
    def token(self):
        if not self._token:
            self._read()
        return self._token

    @token.setter
    def token(self, token):
        super(type(self), type(self)).token.__set__(self, token)  # https://bugs.python.org/issue14965
        self._write()

    @property
    def setTime(self):
        self.token  # Implicitly reads from the file if necessary
        return self._setTime

    def reset(self):
        super().reset()
        with self._lock:
            _logger.info(f'Deleting guest token file {self._file}')
            try:
                os.remove(self._file)
            except FileNotFoundError:
                # Another process likely already removed the file
                pass


class _TwitterTLSAdapter(requests.adapters.HTTPAdapter):
    def init_poolmanager(self, *args, **kwargs):
        # FIXME: When urllib3 2.0.0 is out and can be required, this should use urllib3.util.create_urllib3_context instead of the private, undocumented ssl_ module.
        kwargs['ssl_context'] = urllib3.util.ssl_.create_urllib3_context(ciphers=_CIPHERS_CHROME)
        return super().init_poolmanager(*args, **kwargs)


class _TwitterAPIType(enum.Enum):
    V2 = 0  # Introduced with the redesign
    GRAPHQL = 1


class _TwitterAPIScraper(umbraSnscrpe.base.Scraper):
    def __init__(self, baseUrl, *, guestTokenManager=None, maxEmptyPages=0, **kwargs):
        super().__init__(**kwargs)
        self._baseUrl = baseUrl
        if guestTokenManager is None:
            global _globalGuestTokenManager
            if _globalGuestTokenManager is None:
                _globalGuestTokenManager = GuestTokenManager()
            guestTokenManager = _globalGuestTokenManager
        self._guestTokenManager = guestTokenManager
        self._maxEmptyPages = maxEmptyPages
        self._apiHeaders = {
            'User-Agent': None,
            'Authorization': _API_AUTHORIZATION_HEADER,
            'Referer': self._baseUrl,
            'Accept-Language': 'en-US,en;q=0.5',
        }
        adapter = _TwitterTLSAdapter()
        self._session.mount('https://twitter.com', adapter)
        self._session.mount('https://api.twitter.com', adapter)
        self._set_random_user_agent()

    def _set_random_user_agent(self):
        self._userAgent = f'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.{random.randint(0, 9999)} Safari/537.{random.randint(0, 99)}'
        self._apiHeaders['User-Agent'] = self._userAgent

    def _check_guest_token_response(self, r):
        if r.status_code != 200:
            self._set_random_user_agent()
            return False, f'non-200 response ({r.status_code})'
        return True, None

    def _ensure_guest_token(self, url=None):
        if self._guestTokenManager.token is None:
            _logger.info('Retrieving guest token')
            r = self._get(self._baseUrl if url is None else url, headers={'User-Agent': self._userAgent},
                          responseOkCallback=self._check_guest_token_response)
            if (match := re.search(
                    r'document\.cookie = decodeURIComponent\("gt=(\d+); Max-Age=10800; Domain=\.twitter\.com; Path=/; Secure"\);',
                    r.text)):
                _logger.debug('Found guest token in HTML')
                self._guestTokenManager.token = match.group(1)
            if 'gt' in r.cookies:
                _logger.debug('Found guest token in cookies')
                self._guestTokenManager.token = r.cookies['gt']
            if not self._guestTokenManager.token:
                _logger.debug('No guest token in response')
                _logger.info('Retrieving guest token via API')
                r = self._post('https://api.twitter.com/1.1/guest/activate.json', data=b'', headers=self._apiHeaders,
                               responseOkCallback=self._check_guest_token_response)
                o = r.json()
                if not o.get('guest_token'):
                    raise umbraSnscrpe.base.ScraperException('Unable to retrieve guest token')
                self._guestTokenManager.token = o['guest_token']
            assert self._guestTokenManager.token
        _logger.debug(f'Using guest token {self._guestTokenManager.token}')
        self._session.cookies.set('gt', self._guestTokenManager.token, domain='.twitter.com', path='/', secure=True,
                                  expires=self._guestTokenManager.setTime + _GUEST_TOKEN_VALIDITY)
        self._apiHeaders['x-guest-token'] = self._guestTokenManager.token

    def _unset_guest_token(self):
        self._guestTokenManager.reset()
        del self._session.cookies['gt']
        del self._apiHeaders['x-guest-token']

    def _check_api_response(self, r):
        if r.status_code in (403, 429):
            self._unset_guest_token()
            self._ensure_guest_token()
            return False, f'blocked ({r.status_code})'
        if r.headers.get('content-type', '').replace(' ', '') != 'application/json;charset=utf-8':
            return False, 'content type is not JSON'
        if r.status_code != 200:
            return False, f'non-200 status code ({r.status_code})'
        return True, None

    def _get_api_data(self, endpoint, apiType, params):
        self._ensure_guest_token()
        if apiType is _TwitterAPIType.GRAPHQL:
            params = urllib.parse.urlencode({k: json.dumps(v, separators=(',', ':')) for k, v in params.items()},
                                            quote_via=urllib.parse.quote)
        r = self._get(endpoint, params=params, headers=self._apiHeaders, responseOkCallback=self._check_api_response)
        try:
            obj = r.json()
        except json.JSONDecodeError as e:
            raise umbraSnscrpe.base.ScraperException('Received invalid JSON from Twitter') from e
        return obj

    def _iter_api_data(self, endpoint, apiType, params, paginationParams=None, cursor=None,
                       direction=_ScrollDirection.BOTTOM):
        # Iterate over endpoint with params/paginationParams, optionally starting from a cursor
        # Handles guest token extraction using the baseUrl passed to __init__ etc.
        # Order from params and paginationParams is preserved. To insert the cursor at a particular location, insert a 'cursor' key into paginationParams there (value is overwritten).
        # direction controls in which direction it should scroll from the initial response. BOTH equals TOP followed by BOTTOM.

        # Logic for dual scrolling: direction is set to top, but if the bottom cursor is found, bottomCursorAndStop is set accordingly.
        # Once the top pagination is exhausted, the bottomCursorAndStop is used and reset to None; it isn't set anymore after because the first entry condition will always be true for the bottom cursor.

        if cursor is None:
            reqParams = params
        else:
            reqParams = copy.deepcopy(paginationParams)
            if apiType is _TwitterAPIType.V2:
                reqParams['cursor'] = cursor
            else:
                reqParams['variables']['cursor'] = cursor
        bottomCursorAndStop = None
        if direction is _ScrollDirection.TOP or direction is _ScrollDirection.BOTH:
            dir = 'top'
        else:
            dir = 'bottom'
        stopOnEmptyResponse = False
        emptyResponsesOnCursor = 0
        emptyPages = 0
        while True:
            _logger.info(f'Retrieving scroll page {cursor}')
            obj = self._get_api_data(endpoint, apiType, reqParams)
            yield obj

            # No data format test, just a hard and loud crash if anything's wrong :-)
            newCursor = None
            promptCursor = None
            newBottomCursorAndStop = None
            if apiType is _TwitterAPIType.V2:
                instructions = obj['timeline']['instructions']
            elif apiType is _TwitterAPIType.GRAPHQL:
                if 'user' in obj['data']:
                    # UserTweets, UserTweetsAndReplies
                    instructions = obj['data']['user']['result']['timeline']['timeline']['instructions']
                else:
                    # TweetDetail
                    instructions = obj['data'].get('threaded_conversation_with_injections', {}).get('instructions', [])
            tweetCount = 0
            for instruction in instructions:
                if 'addEntries' in instruction:
                    entries = instruction['addEntries']['entries']
                elif 'replaceEntry' in instruction:
                    entries = [instruction['replaceEntry']['entry']]
                elif instruction.get('type') == 'TimelineAddEntries':
                    entries = instruction['entries']
                else:
                    continue
                tweetCount += self._count_tweets(entries)
                for entry in entries:
                    if not (entry['entryId'].startswith('sq-cursor-') or entry['entryId'].startswith('cursor-')):
                        continue
                    if apiType is _TwitterAPIType.V2:
                        entryCursor = entry['content']['operation']['cursor']['value']
                        entryCursorStop = entry['content']['operation']['cursor'].get('stopOnEmptyResponse', None)
                    elif apiType is _TwitterAPIType.GRAPHQL:
                        cursorContent = entry['content']
                        while cursorContent.get('itemType') == 'TimelineTimelineItem' or cursorContent.get(
                                'entryType') == 'TimelineTimelineItem':
                            cursorContent = cursorContent['itemContent']
                        entryCursor, entryCursorStop = cursorContent['value'], cursorContent.get('stopOnEmptyResponse',
                                                                                                 None)
                    if entry['entryId'] == f'sq-cursor-{dir}' or entry['entryId'].startswith(f'cursor-{dir}-'):
                        newCursor = entryCursor
                        if entryCursorStop is not None:
                            stopOnEmptyResponse = entryCursorStop
                    elif entry['entryId'].startswith(
                            'cursor-showMoreThreadsPrompt-'):  # E.g. 'offensive' replies button
                        promptCursor = entryCursor
                    elif direction is _ScrollDirection.BOTH and bottomCursorAndStop is None and (
                            entry['entryId'] == 'sq-cursor-bottom' or entry['entryId'].startswith('cursor-bottom-')):
                        newBottomCursorAndStop = (entryCursor, entryCursorStop or False)
            if bottomCursorAndStop is None and newBottomCursorAndStop is not None:
                bottomCursorAndStop = newBottomCursorAndStop
            if newCursor == cursor and tweetCount == 0:
                # Twitter sometimes returns the same cursor as requested and no results even though there are more results.
                # When this happens, retry the same cursor up to the retries setting.
                emptyResponsesOnCursor += 1
                if emptyResponsesOnCursor > self._retries:
                    break
            if tweetCount == 0:
                emptyPages += 1
                if self._maxEmptyPages and emptyPages >= self._maxEmptyPages:
                    _logger.warning(f'Stopping after {emptyPages} empty pages')
                    break
            else:
                emptyPages = 0
            if not newCursor or (stopOnEmptyResponse and tweetCount == 0):
                # End of pagination
                if promptCursor is not None:
                    newCursor = promptCursor
                elif direction is _ScrollDirection.BOTH and bottomCursorAndStop is not None:
                    dir = 'bottom'
                    newCursor, stopOnEmptyResponse = bottomCursorAndStop
                    bottomCursorAndStop = None
                else:
                    break
            if newCursor != cursor:
                emptyResponsesOnCursor = 0
            cursor = newCursor
            reqParams = copy.deepcopy(paginationParams)
            if apiType is _TwitterAPIType.V2:
                reqParams['cursor'] = cursor
            else:
                reqParams['variables']['cursor'] = cursor
    def _instructions_to_users(self, obj, includeConversationThreads=False):
        for instruction in obj['globalObjects']['users']:
            yield self._user_to_user(obj['globalObjects']['users'][instruction])

    def _count_tweets(self, entries):
        return sum(entry['entryId'].startswith('sq-I-t-') or entry['entryId'].startswith('tweet-') for entry in entries)

    def _v2_timeline_instructions_to_tweets(self, obj, includeConversationThreads=False):
        # No data format test, just a hard and loud crash if anything's wrong :-)
        for instruction in obj['timeline']['instructions']:
            if 'addEntries' in instruction:
                entries = instruction['addEntries']['entries']
            elif 'replaceEntry' in instruction:
                entries = [instruction['replaceEntry']['entry']]
            else:
                continue
            for entry in entries:
                if entry['entryId'].startswith('sq-I-t-') or entry['entryId'].startswith('tweet-'):
                    yield from self._v2_instruction_tweet_entry_to_tweet(entry['entryId'], entry['content'], obj)
                elif includeConversationThreads and entry['entryId'].startswith('conversationThread-') and not entry[
                    'entryId'].endswith('-show_more_cursor'):
                    for item in entry['content']['timelineModule']['items']:
                        if item['entryId'].startswith('tweet-'):
                            yield from self._v2_instruction_tweet_entry_to_tweet(item['entryId'], item, obj)

    def _v2_instruction_tweet_entry_to_tweet(self, entryId, entry, obj):
        if 'tweet' in entry['item']['content']:
            if 'promotedMetadata' in entry['item']['content']['tweet']:  # Promoted tweet aka ads
                return
            if entry['item']['content']['tweet']['id'] not in obj['globalObjects']['tweets']:
                _logger.warning(
                    f'Skipping tweet {entry["item"]["content"]["tweet"]["id"]} which is not in globalObjects')
                return
            tweet = obj['globalObjects']['tweets'][entry['item']['content']['tweet']['id']]
        elif 'tombstone' in entry['item']['content']:
            if 'tweet' not in entry['item']['content']['tombstone']:  # E.g. deleted reply
                return
            if entry['item']['content']['tombstone']['tweet']['id'] not in obj['globalObjects']['tweets']:
                _logger.warning(
                    f'Skipping tweet {entry["item"]["content"]["tombstone"]["tweet"]["id"]} which is not in globalObjects')
                return
            tweet = obj['globalObjects']['tweets'][entry['item']['content']['tombstone']['tweet']['id']]
        else:
            raise umbraSnscrpe.base.ScraperException(f'Unable to handle entry {entryId!r}')
        yield self._tweet_to_tweet(tweet, obj)

    def _get_tweet_id(self, tweet):
        return tweet['id'] if 'id' in tweet else int(tweet['id_str'])

    def _make_tweet(self, tweet, user, retweetedTweet=None, quotedTweet=None, card=None, **kwargs):
        tweetId = self._get_tweet_id(tweet)
        kwargs['id'] = tweetId
        kwargs['rawContent'] = tweet['full_text']
        kwargs['renderedContent'] = self._render_text_with_urls(tweet['full_text'], tweet['entities'].get('urls'))
        kwargs['user'] = user
        kwargs['date'] = email.utils.parsedate_to_datetime(tweet['created_at'])
        if tweet['entities'].get('urls'):
            kwargs['links'] = [TextLink(
                text=u.get('display_url'),
                url=u['expanded_url'],
                tcourl=u['url'],
                indices=tuple(u['indices']),
            ) for u in tweet['entities']['urls']]
        kwargs['url'] = f'https://twitter.com/{user.username}/status/{tweetId}'
        kwargs['replyCount'] = tweet['reply_count']
        kwargs['retweetCount'] = tweet['retweet_count']
        kwargs['likeCount'] = tweet['favorite_count']
        kwargs['quoteCount'] = tweet['quote_count']
        kwargs['conversationId'] = tweet['conversation_id'] if 'conversation_id' in tweet else int(
            tweet['conversation_id_str'])
        kwargs['lang'] = tweet['lang']
        if 'source' in tweet:
            kwargs['source'] = tweet['source']
            if (match := re.search(r'href=[\'"]?([^\'" >]+)', tweet['source'])):
                kwargs['sourceUrl'] = match.group(1)
            if (match := re.search(r'>([^<]*)<', tweet['source'])):
                kwargs['sourceLabel'] = match.group(1)
        if 'extended_entities' in tweet and 'media' in tweet['extended_entities']:
            media = []
            for medium in tweet['extended_entities']['media']:
                if (mediumO := self._make_medium(medium, tweetId)):
                    media.append(mediumO)
            if media:
                kwargs['media'] = media
        if retweetedTweet:
            kwargs['retweetedTweet'] = retweetedTweet
        if quotedTweet:
            kwargs['quotedTweet'] = quotedTweet
        if (inReplyToTweetId := tweet.get('in_reply_to_status_id_str')):
            kwargs['inReplyToTweetId'] = int(inReplyToTweetId)
            inReplyToUserId = int(tweet['in_reply_to_user_id_str'])
            if inReplyToUserId == kwargs['user'].id:
                kwargs['inReplyToUser'] = kwargs['user']
            elif tweet['entities'].get('user_mentions'):
                for u in tweet['entities']['user_mentions']:
                    if u['id_str'] == tweet['in_reply_to_user_id_str']:
                        kwargs['inReplyToUser'] = User(username=u['screen_name'],
                                                       id=u['id'] if 'id' in u else int(u['id_str']),
                                                       displayname=u['name'])
            if 'inReplyToUser' not in kwargs:
                kwargs['inReplyToUser'] = User(username=tweet['in_reply_to_screen_name'], id=inReplyToUserId)
        if tweet['entities'].get('user_mentions'):
            kwargs['mentionedUsers'] = [
                User(username=u['screen_name'], id=u['id'] if 'id' in u else int(u['id_str']), displayname=u['name'])
                for u in tweet['entities']['user_mentions']]

        # https://developer.twitter.com/en/docs/tutorials/filtering-tweets-by-location
        if tweet.get('coordinates'):
            # coordinates root key (if present) presents coordinates in the form [LONGITUDE, LATITUDE]
            if (coords := tweet['coordinates']['coordinates']) and len(coords) == 2:
                kwargs['coordinates'] = Coordinates(coords[0], coords[1])
        elif tweet.get('geo'):
            # coordinates root key (if present) presents coordinates in the form [LATITUDE, LONGITUDE]
            if (coords := tweet['geo']['coordinates']) and len(coords) == 2:
                kwargs['coordinates'] = Coordinates(coords[1], coords[0])
        if tweet.get('place'):
            kwargs['place'] = Place(tweet['place']['id'], tweet['place']['full_name'], tweet['place']['name'],
                                    tweet['place']['place_type'], tweet['place']['country'],
                                    tweet['place']['country_code'])
            if 'coordinates' not in kwargs and tweet['place'].get('bounding_box') and (
                    coords := tweet['place']['bounding_box']['coordinates']) and coords[0] and len(coords[0][0]) == 2:
                # Take the first (longitude, latitude) couple of the "place square"
                kwargs['coordinates'] = Coordinates(coords[0][0][0], coords[0][0][1])
        if tweet['entities'].get('hashtags'):
            kwargs['hashtags'] = [o['text'] for o in tweet['entities']['hashtags']]
        if tweet['entities'].get('symbols'):
            kwargs['cashtags'] = [o['text'] for o in tweet['entities']['symbols']]
        if card:
            kwargs['card'] = card
            if hasattr(card, 'url') and '//t.co/' in card.url:
                # Try to convert the URL to the non-shortened/t.co one
                # Retweets inherit the card but not the outlinks; try to get them from the retweeted tweet instead in that case.
                candidates = []
                if 'links' in kwargs:
                    candidates.extend(kwargs['links'])
                if retweetedTweet:
                    candidates.extend(retweetedTweet.links)
                for u in candidates:
                    if u.tcourl == card.url:
                        card.url = u.url
                        break
                else:
                    _logger.warning(f'Could not translate t.co card URL on tweet {tweetId}')
        return Tweet(**kwargs)

    def _make_medium(self, medium, tweetId):
        if medium['type'] == 'photo':
            if '?format=' in medium['media_url_https'] or '&format=' in medium['media_url_https']:
                return Photo(previewUrl=medium['media_url_https'], fullUrl=medium['media_url_https'])
            if '.' not in medium['media_url_https']:
                _logger.warning(
                    f'Skipping malformed medium URL on tweet {tweetId}: {medium["media_url_https"]!r} contains no dot')
                return
            baseUrl, format = medium['media_url_https'].rsplit('.', 1)
            if format not in ('jpg', 'png'):
                _logger.warning(f'Skipping photo with unknown format on tweet {tweetId}: {format!r}')
                return
            mKwargs = {
                'previewUrl': f'{baseUrl}?format={format}&name=small',
                'fullUrl': f'{baseUrl}?format={format}&name=large',
            }
            if medium.get('ext_alt_text'):
                mKwargs['altText'] = medium['ext_alt_text']
            return Photo(**mKwargs)
        elif medium['type'] == 'video' or medium['type'] == 'animated_gif':
            variants = []
            for variant in medium['video_info']['variants']:
                variants.append(VideoVariant(contentType=variant['content_type'], url=variant['url'],
                                             bitrate=variant.get('bitrate')))
            mKwargs = {
                'thumbnailUrl': medium['media_url_https'],
                'variants': variants,
            }
            if medium['type'] == 'video':
                mKwargs['duration'] = medium['video_info']['duration_millis'] / 1000
                if (ext := medium.get('ext')) and (mediaStats := ext.get('mediaStats')) and isinstance(
                        r := mediaStats['r'], dict) and 'ok' in r and isinstance(r['ok'], dict):
                    mKwargs['views'] = int(r['ok']['viewCount'])
                elif (mediaStats := medium.get('mediaStats')):
                    mKwargs['views'] = mediaStats['viewCount']
                cls = Video
            elif medium['type'] == 'animated_gif':
                cls = Gif
            if medium.get('ext_alt_text'):
                mKwargs['altText'] = medium['ext_alt_text']
            return cls(**mKwargs)
        else:
            _logger.warning(f'Unsupported medium type on tweet {tweetId}: {medium["type"]!r}')

    def _make_card(self, card, apiType, tweetId):
        bindingValues = {}

        def _kwargs_from_map(keyKwargMap):
            nonlocal bindingValues
            return {kwarg: bindingValues[key] for key, kwarg in keyKwargMap.items() if key in bindingValues}

        userRefs = {}
        if apiType is _TwitterAPIType.V2:
            for o in card.get('users', {}).values():
                userId = o['id']
                assert userId not in userRefs
                userRefs[userId] = self._user_to_user(o)
        elif apiType is _TwitterAPIType.GRAPHQL:
            for o in card['legacy'].get('user_refs_results', []):
                o = o['result']
                if o['__typename'] == 'UserUnavailable':
                    _logger.warning(f'Unavailable user in card on tweet {tweetId}')
                    continue
                userId = int(o['rest_id'])
                if 'legacy' in o:
                    user = self._user_to_user(o['legacy'], id_=userId)
                else:
                    user = UserRef(id=userId)
                if userId in userRefs:
                    if userRefs[userId] != user:
                        _logger.warning(f'Duplicate user {userId} with differing data in card on tweet {tweetId}')
                    continue
                userRefs[userId] = user

        if apiType is _TwitterAPIType.V2:
            messyBindingValues = card['binding_values'].items()
        elif apiType is _TwitterAPIType.GRAPHQL:
            messyBindingValues = ((x['key'], x['value']) for x in card['legacy']['binding_values'])
        for key, value in messyBindingValues:
            if 'type' not in value:
                # Silently ignore creator/site entries since they frequently appear like this.
                if key not in ('creator', 'site'):
                    _logger.warning(f'Skipping type-less card value {key!r} on tweet {tweetId}')
                continue
            if value['type'] == 'STRING':
                bindingValues[key] = value['string_value']
                if key.endswith('_datetime_utc'):
                    bindingValues[key] = datetime.datetime.strptime(bindingValues[key], '%Y-%m-%dT%H:%M:%SZ').replace(
                        tzinfo=datetime.timezone.utc)
            elif value['type'] == 'IMAGE':
                bindingValues[key] = value['image_value']['url']
            elif value['type'] == 'IMAGE_COLOR':
                # Silently discard this.
                pass
            elif value['type'] == 'BOOLEAN':
                bindingValues[key] = value['boolean_value']
            elif value['type'] == 'USER':
                userId = int(value['user_value']['id_str'])
                bindingValues[key] = userRefs.get(userId)
                if bindingValues[key] is None:
                    _logger.warning(f'User {userId} not found in user refs in card on tweet {tweetId}')
            else:
                _logger.warning(f'Unsupported card value type on {key!r} on tweet {tweetId}: {value["type"]!r}')

        if apiType is _TwitterAPIType.V2:
            cardName = card['name']
        elif apiType is _TwitterAPIType.GRAPHQL:
            cardName = card['legacy']['name']

        if cardName in ('summary', 'summary_large_image', 'app', 'direct_store_link_app'):
            keyKwargMap = {
                'title': 'title',
                'description': 'description',
                'card_url': 'url',
                'site': 'siteUser',
                'creator': 'creatorUser',
            }
            if cardName in ('app', 'direct_store_link_app'):
                keyKwargMap['thumbnail_original'] = 'thumbnailUrl'
                return AppCard(**_kwargs_from_map(keyKwargMap))
            else:
                keyKwargMap['thumbnail_image_original'] = 'thumbnailUrl'
                return SummaryCard(**_kwargs_from_map(keyKwargMap))
        elif any(cardName.startswith(x) for x in ('poll2choice_', 'poll3choice_', 'poll4choice_')) and \
                cardName.split('_', 1)[1] in ('text_only', 'image', 'video'):
            kwargs = _kwargs_from_map({'end_datetime_utc': 'endDate', 'last_updated_datetime_utc': 'lastUpdateDate',
                                       'duration_minutes': 'duration', 'counts_are_final': 'finalResults'})

            options = []
            for key in sorted(bindingValues):
                if key.startswith('choice') and key.endswith('_label'):
                    optKwargs = {'label': bindingValues[key]}
                    if (count := bindingValues.get(f'{key[:-5]}count')):
                        optKwargs['count'] = int(count)
                    options.append(PollOption(**optKwargs))
            kwargs['options'] = options
            kwargs['duration'] = int(kwargs['duration'])

            if cardName.endswith('_image'):
                kwargs['medium'] = Photo(previewUrl=bindingValues['image_small'],
                                         fullUrl=bindingValues['image_original'])
            elif cardName.endswith('_video'):
                variants = []
                variants.append(VideoVariant(contentType='application/x-mpegurl', url=bindingValues['player_hls_url'],
                                             bitrate=None))
                if 'vmap' not in bindingValues['player_stream_url']:
                    _logger.warning(f'Non-VMAP URL in {cardName} player_stream_url on tweet {tweetId}')
                variants.append(
                    VideoVariant(contentType='text/xml', url=bindingValues['player_stream_url'], bitrate=None))
                kwargs['medium'] = Video(thumbnailUrl=bindingValues['player_image_original'], variants=variants,
                                         duration=int(bindingValues['content_duration_seconds']))

            return PollCard(**kwargs)
        elif cardName == 'player':
            return PlayerCard(**_kwargs_from_map(
                {'title': 'title', 'description': 'description', 'card_url': 'url', 'player_image_original': 'imageUrl',
                 'site': 'siteUser'}))
        elif cardName in ('promo_image_convo', 'promo_video_convo'):
            kwargs = _kwargs_from_map({'thank_you_text': 'thankYouText', 'thank_you_url': 'thankYouUrl',
                                       'thank_you_shortened_url': 'thankYouTcoUrl'})
            kwargs['actions'] = []
            for l in ('one', 'two', 'three', 'four'):
                if f'cta_{l}' in bindingValues:
                    kwargs['actions'].append(
                        PromoConvoAction(label=bindingValues[f'cta_{l}'], tweet=bindingValues[f'cta_{l}_tweet']))
            if 'image' in cardName:
                kwargs['medium'] = Photo(previewUrl=bindingValues['promo_image_small'],
                                         fullUrl=bindingValues['promo_image_original'])
                if 'cover_promo_image' in bindingValues:
                    kwargs['cover'] = Photo(previewUrl=bindingValues['cover_promo_image_small'],
                                            fullUrl=bindingValues['cover_promo_image_original'])
            elif 'video' in cardName:
                variants = []
                variants.append(VideoVariant(contentType=bindingValues['player_stream_content_type'],
                                             url=bindingValues['player_stream_url'], bitrate=None))
                if bindingValues['player_stream_url'] != bindingValues['player_url']:
                    if 'vmap' not in bindingValues['player_url']:
                        _logger.warning(f'Non-VMAP URL in {cardName} player_url on tweet {tweetId}')
                    variants.append(VideoVariant(contentType='text/xml', url=bindingValues['player_url'], bitrate=None))
                kwargs['medium'] = Video(thumbnailUrl=bindingValues['player_image_original'], variants=variants,
                                         duration=int(bindingValues['content_duration_seconds']))
            return PromoConvoCard(**kwargs)
        elif cardName in ('745291183405076480:broadcast', '3691233323:periscope_broadcast'):
            keyKwargMap = {'broadcast_state': 'state', 'broadcast_source': 'source', 'site': 'siteUser'}
            if cardName == '745291183405076480:broadcast':
                keyKwargMap = {**keyKwargMap, 'broadcast_id': 'id', 'broadcast_url': 'url', 'broadcast_title': 'title',
                               'broadcast_thumbnail_original': 'thumbnailUrl'}
            else:
                keyKwargMap = {**keyKwargMap, 'id': 'id', 'url': 'url', 'title': 'title', 'description': 'description',
                               'total_participants': 'totalParticipants', 'full_size_thumbnail_url': 'thumbnailUrl'}
            kwargs = _kwargs_from_map(keyKwargMap)
            if 'broadcaster_twitter_id' in bindingValues:
                kwargs['broadcaster'] = User(id=int(bindingValues['broadcaster_twitter_id']),
                                             username=bindingValues['broadcaster_username'],
                                             displayname=bindingValues['broadcaster_display_name'])
            if 'siteUser' not in kwargs:
                kwargs['siteUser'] = None
            if cardName == '745291183405076480:broadcast':
                return BroadcastCard(**kwargs)
            else:
                kwargs['totalParticipants'] = int(kwargs['totalParticipants'])
                return PeriscopeBroadcastCard(**kwargs)
        elif cardName == '745291183405076480:live_event':
            kwargs = _kwargs_from_map({'event_id': 'id', 'event_title': 'title', 'event_category': 'category',
                                       'event_subtitle': 'description'})
            kwargs['id'] = int(kwargs['id'])
            kwargs['photo'] = Photo(previewUrl=bindingValues['event_thumbnail_small'],
                                    fullUrl=bindingValues.get('event_thumbnail_original') or bindingValues[
                                        'event_thumbnail'])
            return EventCard(event=Event(**kwargs))
        elif cardName == '3337203208:newsletter_publication':
            kwargs = _kwargs_from_map({'newsletter_title': 'title', 'newsletter_description': 'description',
                                       'newsletter_image_original': 'imageUrl', 'card_url': 'url',
                                       'revue_account_id': 'revueAccountId', 'issue_count': 'issueCount'})
            kwargs['revueAccountId'] = int(kwargs['revueAccountId'])
            kwargs['issueCount'] = int(kwargs['issueCount'])
            return NewsletterCard(**kwargs)
        elif cardName == '3337203208:newsletter_issue':
            kwargs = _kwargs_from_map({
                'newsletter_title': 'newsletterTitle',
                'newsletter_description': 'newsletterDescription',
                'issue_title': 'issueTitle',
                'issue_description': 'issueDescription',
                'issue_number': 'issueNumber',
                'issue_image_original': 'imageUrl',
                'card_url': 'url',
                'revue_account_id': 'revueAccountId'
            })
            kwargs['issueNumber'] = int(kwargs['issueNumber'])
            kwargs['revueAccountId'] = int(kwargs['revueAccountId'])
            return NewsletterIssueCard(**kwargs)
        elif cardName == 'amplify':
            return AmplifyCard(
                id=bindingValues['amplify_content_id'],
                video=Video(
                    thumbnailUrl=bindingValues['player_image'],
                    variants=[VideoVariant(url=bindingValues['amplify_url_vmap'],
                                           contentType=bindingValues.get('player_stream_content_type'), bitrate=None)],
                ),
            )
        elif cardName == 'appplayer':
            kwargs = _kwargs_from_map(
                {'title': 'title', 'app_category': 'appCategory', 'player_owner_id': 'playerOwnerId',
                 'site': 'siteUser'})
            kwargs['playerOwnerId'] = int(kwargs['playerOwnerId'])
            variants = []
            variants.append(
                VideoVariant(contentType='application/x-mpegurl', url=bindingValues['player_hls_url'], bitrate=None))
            if 'vmap' not in bindingValues['player_url']:
                _logger.warning(f'Non-VMAP URL in {cardName} player_url on tweet {tweetId}')
            variants.append(VideoVariant(contentType='text/xml', url=bindingValues['player_url'], bitrate=None))
            kwargs['video'] = Video(thumbnailUrl=bindingValues['player_image_original'], variants=variants,
                                    duration=int(bindingValues['content_duration_seconds']))
            return AppPlayerCard(**kwargs)
        elif cardName == '3691233323:audiospace':
            return SpacesCard(**_kwargs_from_map({'card_url': 'url', 'id': 'id'}))
        elif cardName == '2586390716:message_me':
            # Note that the strings in Twitter's JS appear to have an incorrect mapping that then gets changed somewhere in the 1.8 MiB of JS!
            # cta_1, 3, and 4 should mean 'Message us', 'Send a private message', and 'Send me a private message', but the correct mapping is currently unknown.
            ctas = {'message_me_card_cta_2': 'Send us a private message'}
            if bindingValues['cta'] not in ctas:
                _logger.warning(f'Unsupported message_me card cta on tweet {tweetId}: {bindingValues["cta"]!r}')
                return
            return MessageMeCard(**_kwargs_from_map({'recipient': 'recipient', 'card_url': 'url'}),
                                 buttonText=ctas[bindingValues['cta']])
        elif cardName == 'unified_card':
            o = json.loads(bindingValues['unified_card'])
            kwargs = {}
            if 'type' in o:
                unifiedCardType = o.get('type')
                if unifiedCardType not in (
                        'image_app',
                        'image_carousel_app',
                        'image_carousel_website',
                        'image_collection_website',
                        'image_multi_dest_carousel_website',
                        'image_website',
                        'mixed_media_multi_dest_carousel_website',
                        'mixed_media_single_dest_carousel_app',
                        'mixed_media_single_dest_carousel_website',
                        'video_app',
                        'video_carousel_app',
                        'video_carousel_website',
                        'video_multi_dest_carousel_website',
                        'video_website',
                ):
                    _logger.warning(f'Unsupported unified_card type on tweet {tweetId}: {unifiedCardType!r}')
                    return
                kwargs['type'] = unifiedCardType
            elif set(c['type'] for c in o['component_objects'].values()) not in (
                    {'media', 'twitter_list_details'}, {'media', 'community_details'}):
                _logger.warning(f'Unsupported unified_card type on tweet {tweetId}')
                return

            kwargs['componentObjects'] = {}
            for k, v in o['component_objects'].items():
                if v['type'] == 'details':
                    co = UnifiedCardDetailComponentObject(content=v['data']['title']['content'],
                                                          destinationKey=v['data']['destination'])
                elif v['type'] == 'media':
                    co = UnifiedCardMediumComponentObject(mediumKey=v['data']['id'],
                                                          destinationKey=v['data']['destination'])
                elif v['type'] == 'button_group':
                    if not all(b['type'] == 'cta' for b in v['data']['buttons']):
                        _logger.warning(f'Unsupported unified_card button_group button type on tweet {tweetId}')
                        return
                    buttons = [UnifiedCardButton(
                        text=b['action'][0].upper() + re.sub('[A-Z]', lambda x: f' {x[0]}', b['action'][1:]),
                        destinationKey=b['destination']) for b in v['data']['buttons']]
                    co = UnifiedCardButtonGroupComponentObject(buttons=buttons)
                elif v['type'] == 'swipeable_media':
                    media = [UnifiedCardSwipeableMediaMedium(mediumKey=m['id'], destinationKey=m['destination']) for m
                             in v['data']['media_list']]
                    co = UnifiedCardSwipeableMediaComponentObject(media=media)
                elif v['type'] == 'app_store_details':
                    co = UnifiedCardAppStoreComponentObject(appKey=v['data']['app_id'],
                                                            destinationKey=v['data']['destination'])
                elif v['type'] == 'twitter_list_details':
                    co = UnifiedCardTwitterListDetailsComponentObject(
                        name=v['data']['name']['content'],
                        memberCount=v['data']['member_count'],
                        subscriberCount=v['data']['subscriber_count'],
                        user=self._user_to_user(o['users'][v['data']['user_id']]),
                        destinationKey=v['data']['destination'],
                    )
                elif v['type'] == 'community_details':
                    co = UnifiedCardTwitterCommunityDetailsComponentObject(
                        name=v['data']['name']['content'],
                        theme=v['data']['theme'],
                        membersCount=v['data']['member_count'],
                        destinationKey=v['data']['destination'],
                        membersFacepile=[self._user_to_user(u) for u in
                                         map(o['users'].get, v['data']['members_facepile']) if u],
                    )
                else:
                    _logger.warning(f'Unsupported unified_card component type on tweet {tweetId}: {v["type"]!r}')
                    return
                kwargs['componentObjects'][k] = co

            kwargs['destinations'] = {}
            for k, v in o['destination_objects'].items():
                dKwargs = {}
                if 'url_data' in v['data']:
                    dKwargs['url'] = v['data']['url_data']['url']
                if 'app_id' in v['data']:
                    dKwargs['appKey'] = v['data']['app_id']
                if 'media_id' in v['data']:
                    dKwargs['mediumKey'] = v['data']['media_id']
                kwargs['destinations'][k] = UnifiedCardDestination(**dKwargs)

            kwargs['media'] = {}
            for k, v in o['media_entities'].items():
                if (medium := self._make_medium(v, tweetId)):
                    kwargs['media'][k] = medium

            if 'app_store_data' in o:
                kwargs['apps'] = {}
                for k, v in o['app_store_data'].items():
                    variants = []
                    for var in v:
                        vKwargsMap = {
                            'type': 'type',
                            'id': 'id',
                            'icon_media_key': 'iconMediumKey',
                            'country_code': 'countryCode',
                            'num_installs': 'installs',
                            'size_bytes': 'size',
                            'is_free': 'isFree',
                            'is_editors_choice': 'isEditorsChoice',
                            'has_in_app_purchases': 'hasInAppPurchases',
                            'has_in_app_ads': 'hasInAppAds',
                        }
                        vKwargs = {kwarg: var[key] for key, kwarg in vKwargsMap.items() if key in var}
                        vKwargs['title'] = var['title']['content']
                        if 'description' in var:
                            vKwargs['description'] = var['description']['content']
                        if 'category' in var:
                            vKwargs['category'] = var['category']['content']
                        if (ratings := var['ratings']):
                            vKwargs['ratingAverage'] = var['ratings']['star']
                            vKwargs['ratingCount'] = var['ratings']['count']
                        vKwargs['url'] = f'https://play.google.com/store/apps/details?id={var["id"]}' if var[
                                                                                                             'type'] == 'android_app' else f'https://itunes.apple.com/app/id{var["id"]}'
                        if 'iconMediumKey' in vKwargs and vKwargs['iconMediumKey'] not in kwargs['media']:
                            # https://github.com/JustAnotherArchivist/snscrape/issues/470
                            _logger.warning(
                                f'Tweet {tweetId} contains an app icon medium key {vKwargs["iconMediumKey"]!r} on app {vKwargs["type"]!r}/{vKwargs["id"]!r}, but the corresponding medium is missing; dropping')
                            del vKwargs['iconMediumKey']
                        variants.append(UnifiedCardApp(**vKwargs))
                    kwargs['apps'][k] = variants

            if o['components']:
                kwargs['components'] = o['components']

            if 'layout' in o:
                if o['layout']['type'] == 'swipeable':
                    kwargs['swipeableLayoutSlides'] = [
                        UnifiedCardSwipeableLayoutSlide(mediumComponentKey=v[0], componentKey=v[1]) for v in
                        o['layout']['data']['slides']]
                elif o['layout']['type'] == 'collection':
                    kwargs['collectionLayoutSlides'] = [
                        UnifiedCardCollectionLayoutSlide(detailsComponentKey=v[0], mediumComponentKey=v[1]) for v in
                        o['layout']['data']['slides']]
                else:
                    _logger.warning(f'Unsupported unified_card layout type on tweet {tweetId}: {o["layout"]["type"]!r}')
                    return

            return UnifiedCard(**kwargs)

        _logger.warning(f'Unsupported card type on tweet {tweetId}: {cardName!r}')

    def _make_vibe(self, vibe):
        return Vibe(
            text=vibe['text'],
            imageUrl=vibe['imgUrl'],
            imageDescription=vibe['imgDescription'],
        )

    def _tweet_to_tweet(self, tweet, obj):
        user = self._user_to_user(obj['globalObjects']['users'][tweet['user_id_str']])
        kwargs = {}
        if 'retweeted_status_id_str' in tweet:
            kwargs['retweetedTweet'] = self._tweet_to_tweet(
                obj['globalObjects']['tweets'][tweet['retweeted_status_id_str']], obj)
        if 'quoted_status_id_str' in tweet and tweet['quoted_status_id_str'] in obj['globalObjects']['tweets']:
            kwargs['quotedTweet'] = self._tweet_to_tweet(obj['globalObjects']['tweets'][tweet['quoted_status_id_str']],
                                                         obj)
        if 'card' in tweet:
            kwargs['card'] = self._make_card(tweet['card'], _TwitterAPIType.V2, self._get_tweet_id(tweet))
        if 'ext_views' in tweet and 'count' in tweet['ext_views']:
            kwargs['viewCount'] = int(tweet['ext_views']['count'])
        if 'vibe' in tweet.get('ext', {}):
            kwargs['vibe'] = self._make_vibe(tweet['ext']['vibe']['r']['ok'])
        return self._make_tweet(tweet, user, **kwargs)

    def _graphql_timeline_tweet_item_result_to_tweet(self, result):
        if result['__typename'] == 'Tweet':
            pass
        elif result['__typename'] == 'TweetWithVisibilityResults':
            # TODO Include result['softInterventionPivot'] in the Tweet object
            result = result['tweet']
        else:
            raise umbraSnscrpe.base.ScraperException(f'Unknown result type {result["__typename"]!r}')
        tweet = result['legacy']
        userId = int(result['core']['user_results']['result']['rest_id'])
        user = self._user_to_user(result['core']['user_results']['result']['legacy'], id_=userId)
        kwargs = {}
        if 'retweeted_status_result' in tweet:
            kwargs['retweetedTweet'] = self._graphql_timeline_tweet_item_result_to_tweet(
                tweet['retweeted_status_result']['result'])
        if 'quoted_status_result' in result:
            if result['quoted_status_result']['result']['__typename'] == 'TweetTombstone':
                kwargs['quotedTweet'] = TweetRef(id=int(tweet['quoted_status_id_str']))
            else:
                kwargs['quotedTweet'] = self._graphql_timeline_tweet_item_result_to_tweet(
                    result['quoted_status_result']['result'])
        elif 'quotedRefResult' in result:
            if result['quotedRefResult']['result']['__typename'] == 'TweetTombstone':
                kwargs['quotedTweet'] = TweetRef(id=int(tweet['quoted_status_id_str']))
            else:
                kwargs['quotedTweet'] = TweetRef(id=int(result['quotedRefResult']['result']['rest_id']))
        elif 'quoted_status_id_str' in tweet:
            kwargs['quotedTweet'] = TweetRef(id=int(tweet['quoted_status_id_str']))
        if 'card' in result:
            kwargs['card'] = self._make_card(result['card'], _TwitterAPIType.GRAPHQL, self._get_tweet_id(tweet))
        if 'views' in result and 'count' in result['views']:
            kwargs['viewCount'] = int(result['views']['count'])
        if 'vibe' in result:
            kwargs['vibe'] = self._make_vibe(result['vibe'])
        return self._make_tweet(tweet, user, **kwargs)

    def _graphql_timeline_instructions_to_tweets(self, instructions, includeConversationThreads=False):
        for instruction in instructions:
            if instruction['type'] != 'TimelineAddEntries':
                continue
            for entry in instruction['entries']:
                if entry['entryId'].startswith('tweet-'):
                    if entry['content']['entryType'] == 'TimelineTimelineItem' and entry['content']['itemContent'][
                        'itemType'] == 'TimelineTweet':
                        if 'result' not in entry['content']['itemContent']['tweet_results']:
                            _logger.warning(f'Skipping empty tweet entry {entry["entryId"]}')
                            continue
                        yield self._graphql_timeline_tweet_item_result_to_tweet(
                            entry['content']['itemContent']['tweet_results']['result'])
                    else:
                        _logger.warning('Got unrecognised timeline tweet item(s)')
                elif includeConversationThreads and entry['entryId'].startswith(
                        'conversationthread-'):  # TODO show more cursor?
                    for item in entry['content']['items']:
                        if item['entryId'].startswith(f'{entry["entryId"]}-tweet-'):
                            yield self._graphql_timeline_tweet_item_result_to_tweet(
                                item['item']['itemContent']['tweet_results']['result'])

    def _render_text_with_urls(self, text, urls):
        if not urls:
            return text
        out = []
        out.append(text[:urls[0]['indices'][0]])
        urlsSorted = sorted(urls,
                            key=lambda x: x['indices'][0])  # Ensure that they're in left to right appearance order
        assert all(
            url['indices'][1] <= nextUrl['indices'][0] for url, nextUrl in zip(urls, urls[1:])), 'broken URL indices'
        for url, nextUrl in itertools.zip_longest(urls, urls[1:]):
            if 'display_url' in url:
                out.append(url['display_url'])
            out.append(text[url['indices'][1]: nextUrl['indices'][0] if nextUrl is not None else None])
        return ''.join(out)

    def _user_to_user(self, user, id_=None):
        kwargs = {}
        kwargs['username'] = user['screen_name']
        kwargs['id'] = id_ if id_ else user['id'] if 'id' in user else int(user['id_str'])
        kwargs['displayname'] = user['name']
        kwargs['rawDescription'] = user['description']
        kwargs['renderedDescription'] = self._render_text_with_urls(user['description'],
                                                                    user['entities']['description'].get('urls'))
        if user['entities']['description'].get('urls'):
            kwargs['descriptionLinks'] = [TextLink(
                text=x.get('display_url'),
                url=x['expanded_url'],
                tcourl=x['url'],
                indices=tuple(x['indices']),
            ) for x in user['entities']['description']['urls']]
        kwargs['verified'] = user.get('verified')
        kwargs['created'] = email.utils.parsedate_to_datetime(user['created_at'])
        kwargs['followersCount'] = user['followers_count']
        kwargs['friendsCount'] = user['friends_count']
        kwargs['statusesCount'] = user['statuses_count']
        kwargs['favouritesCount'] = user['favourites_count']
        kwargs['listedCount'] = user['listed_count']
        kwargs['mediaCount'] = user['media_count']
        kwargs['location'] = user['location']
        kwargs['protected'] = user.get('protected')
        if user.get('url'):
            entity = user['entities'].get('url', {}).get('urls', [None])[0]
            if not entity or entity['url'] != user['url']:
                self.logger.warning(f'Link inconsistency on user {kwargs["id"]}')
            if not entity:
                entity = {'indices': (0, len(user['url']))}
            kwargs['link'] = TextLink(text=entity.get('display_url'), url=entity.get('expanded_url', user['url']),
                                      tcourl=user['url'], indices=tuple(entity['indices']))
        kwargs['profileImageUrl'] = user['profile_image_url_https']
        kwargs['profileBannerUrl'] = user.get('profile_banner_url')
        if 'ext' in user and 'highlightedLabel' in user['ext'] and (
                label := user['ext']['highlightedLabel']['r']['ok'].get('label')):
            kwargs['label'] = self._user_label_to_user_label(label)
        return User(**kwargs)

    def _user_label_to_user_label(self, label):
        labelKwargs = {}
        labelKwargs['description'] = label['description']
        if 'url' in label and 'url' in label['url']:
            labelKwargs['url'] = label['url']['url']
        if 'badge' in label and 'url' in label['badge']:
            labelKwargs['badgeUrl'] = label['badge']['url']
        if 'longDescription' in label and 'text' in label['longDescription']:
            labelKwargs['longDescription'] = label['longDescription']['text']
        return UserLabel(**labelKwargs)

    @classmethod
    def _cli_construct(cls, argparseArgs, *args, **kwargs):
        kwargs['guestTokenManager'] = _CLIGuestTokenManager()
        return super()._cli_construct(argparseArgs, *args, **kwargs)


class TwitterSearchScraper(_TwitterAPIScraper):
    name = 'twitter-search'

    def __init__(self, query, *, cursor=None, top=False, maxEmptyPages=20, **kwargs):
        if not query.strip():
            raise ValueError('empty query')
        kwargs['maxEmptyPages'] = maxEmptyPages
        super().__init__(baseUrl='https://twitter.com/search?' + urllib.parse.urlencode(
            {'f': 'live', 'lang': 'en', 'q': query, 'src': 'spelling_expansion_revert_click'}), **kwargs)
        self._query = query  # Note: may get replaced by subclasses when using user ID resolution
        self._cursor = cursor
        self._top = top

    def _check_scroll_response(self, r):
        if r.status_code == 429:
            # Accept a 429 response as "valid" to prevent retries; handled explicitly in get_items
            return True, None
        if r.headers.get('content-type').replace(' ', '') != 'application/json;charset=utf-8':
            return False, 'content type is not JSON'
        if r.status_code != 200:
            return False, 'non-200 status code'
        return True, None

    def get_items(self):
        if not self._query.strip():
            raise ValueError('empty query')
        paginationParams = {
            'include_profile_interstitial_type': '1',
            'include_blocking': '1',
            'include_blocked_by': '1',
            'include_followed_by': '1',
            'include_want_retweets': '1',
            'include_mute_edge': '1',
            'include_can_dm': '1',
            'include_can_media_tag': '1',
            'include_ext_has_nft_avatar': '1',
            'include_ext_is_blue_verified': '1',
            'include_ext_verified_type': '1',
            'skip_status': '1',
            'cards_platform': 'Web-12',
            'include_cards': '1',
            'include_ext_alt_text': 'true',
            'include_ext_limited_action_results': 'false',
            'include_quote_count': 'true',
            'include_reply_count': '1',
            'tweet_mode': 'extended',
            'include_ext_collab_control': 'true',
            'include_ext_views': 'true',
            'include_entities': 'true',
            'include_user_entities': 'true',
            'include_ext_media_color': 'true',
            'include_ext_media_availability': 'true',
            'include_ext_sensitive_media_warning': 'true',
            'include_ext_trusted_friends_metadata': 'true',
            'send_error_codes': 'true',
            'simple_quoted_tweet': 'true',
            'q': self._query,
            'tweet_search_mode': 'live',
            'count': '20',
            'query_source': 'spelling_expansion_revert_click',
            'cursor': None,
            'pc': '1',
            'spelling_corrections': '1',
            'include_ext_edit_control': 'true',
            'ext': 'mediaStats,highlightedLabel,hasNftAvatar,voiceInfo,enrichments,superFollowMetadata,unmentionInfo,editControl,collab_control,vibe',
        }
        params = paginationParams.copy()
        del params['cursor']

        if self._top:
            del params['tweet_search_mode']
            del paginationParams['tweet_search_mode']

        for obj in self._iter_api_data('https://api.twitter.com/2/search/adaptive.json', _TwitterAPIType.V2, params,
                                       paginationParams, cursor=self._cursor):
            yield from self._v2_timeline_instructions_to_tweets(obj)

    @classmethod
    def _cli_setup_parser(cls, subparser):
        subparser.add_argument('--cursor', metavar='CURSOR')
        subparser.add_argument('--top', action='store_true', default=False,
                               help='Enable fetching top tweets instead of live/chronological')
        subparser.add_argument('--max-empty-pages', dest='maxEmptyPages', metavar='N', type=int, default=20,
                               help='Stop after N empty pages from Twitter; set to 0 to disable')
        subparser.add_argument('query', type=umbraSnscrpe.base.nonempty_string('query'), help='A Twitter search string')

    @classmethod
    def _cli_from_args(cls, args):
        return cls._cli_construct(args, args.query, cursor=args.cursor, top=args.top, maxEmptyPages=args.maxEmptyPages)


class TwitterSearchScraperPeople(_TwitterAPIScraper):
    name = 'twitter-search-people'

    def __init__(self, query, *, cursor=None, top=False, maxEmptyPages=20, **kwargs):
        if not query.strip():
            raise ValueError('empty query')
        kwargs['maxEmptyPages'] = maxEmptyPages
        super().__init__(baseUrl='https://twitter.com/search?' + urllib.parse.urlencode(
            {'f': 'live', 'lang': 'en', 'q': query, 'src': 'spelling_expansion_revert_click'}), **kwargs)
        self._query = query  # Note: may get replaced by subclasses when using user ID resolution
        self._cursor = cursor
        self._top = top

    def _check_scroll_response(self, r):
        if r.status_code == 429:
            # Accept a 429 response as "valid" to prevent retries; handled explicitly in get_items
            return True, None
        if r.headers.get('content-type').replace(' ', '') != 'application/json;charset=utf-8':
            return False, 'content type is not JSON'
        if r.status_code != 200:
            return False, 'non-200 status code'
        return True, None

    def get_items(self):
        if not self._query.strip():
            raise ValueError('empty query')
        paginationParams = {
            'include_profile_interstitial_type': '1',
            'include_blocking': '1',
            'include_blocked_by': '1',
            'include_followed_by': '1',
            'include_want_retweets': '1',
            'include_mute_edge': '1',
            'include_can_dm': '1',
            'include_can_media_tag': '1',
            'include_ext_has_nft_avatar': '1',
            'include_ext_is_blue_verified': '1',
            'include_ext_verified_type': '1',
            'skip_status': '1',
            'cards_platform': 'Web-12',
            'include_cards': '1',
            'include_ext_alt_text': 'true',
            'include_ext_limited_action_results': 'false',
            'include_quote_count': 'true',
            'include_reply_count': '1',
            'tweet_mode': 'extended',
            'include_ext_collab_control': 'true',
            'include_ext_views': 'true',
            'include_entities': 'true',
            'include_user_entities': 'true',
            'include_ext_media_color': 'true',
            'include_ext_media_availability': 'true',
            'include_ext_sensitive_media_warning': 'true',
            'include_ext_trusted_friends_metadata': 'true',
            'send_error_codes': 'true',
            'simple_quoted_tweet': 'true',
            'q': self._query,
            # 'tweet_search_mode': 'live',
            'result_filter': 'user',
            'count': '20',
            'query_source': 'spelling_expansion_revert_click',
            'cursor': None,
            'pc': '1',
            'spelling_corrections': '1',
            'include_ext_edit_control': 'true',
            'ext': 'mediaStats,highlightedLabel,hasNftAvatar,voiceInfo,enrichments,superFollowMetadata,unmentionInfo,editControl,collab_control,vibe',
        }
        params = paginationParams.copy()
        del params['cursor']

        if self._top:
            del params['tweet_search_mode']
            del paginationParams['tweet_search_mode']

        for obj in self._iter_api_data('https://api.twitter.com/2/search/adaptive.json', _TwitterAPIType.V2, params,
                                       paginationParams, cursor=self._cursor):
            yield from self._instructions_to_users(obj)

    @classmethod
    def _cli_setup_parser(cls, subparser):
        subparser.add_argument('--cursor', metavar='CURSOR')
        subparser.add_argument('--top', action='store_true', default=False,
                               help='Enable fetching top tweets instead of live/chronological')
        subparser.add_argument('--max-empty-pages', dest='maxEmptyPages', metavar='N', type=int, default=20,
                               help='Stop after N empty pages from Twitter; set to 0 to disable')
        subparser.add_argument('query', type=umbraSnscrpe.base.nonempty_string('query'), help='A Twitter search string')

    @classmethod
    def _cli_from_args(cls, args):
        return cls._cli_construct(args, args.query, cursor=args.cursor, top=args.top, maxEmptyPages=args.maxEmptyPages)


class TwitterUserScraper(TwitterSearchScraper):
    name = 'twitter-user'

    def __init__(self, user, **kwargs):
        self._isUserId = isinstance(user, int)
        if not self._isUserId and not self.is_valid_username(user):
            raise ValueError('Invalid username')
        super().__init__(f'from:{user}', **kwargs)
        self._user = user
        self._baseUrl = f'https://twitter.com/{self._user}' if not self._isUserId else f'https://twitter.com/i/user/{self._user}'

    def _get_entity(self):
        self._ensure_guest_token()
        if not self._isUserId:
            fieldName = 'screen_name'
            endpoint = 'https://twitter.com/i/api/graphql/7mjxD3-C6BxitPMVQ6w0-Q/UserByScreenName'
        else:
            fieldName = 'userId'
            endpoint = 'https://twitter.com/i/api/graphql/I5nvpI91ljifos1Y3Lltyg/UserByRestId'
        variables = {fieldName: str(self._user), 'withSafetyModeUserFields': True, 'withSuperFollowsUserFields': True}
        obj = self._get_api_data(endpoint, _TwitterAPIType.GRAPHQL, params={'variables': variables})
        if not obj['data'] or obj['data']['user']['result']['__typename'] == 'UserUnavailable':
            return None
        user = obj['data']['user']['result']
        rawDescription = user['legacy']['description']
        renderedDescription = self._render_text_with_urls(rawDescription,
                                                          user['legacy']['entities']['description']['urls'])
        link = None
        if user['legacy'].get('url'):
            entity = user['legacy']['entities'].get('url', {}).get('urls', [None])[0]
            if not entity or entity['url'] != user['legacy']['url']:
                self.logger.warning(f'Link inconsistency on user')
            if not entity:
                entity = {'indices': (0, len(user['legacy']['url']))}
            link = TextLink(text=entity.get('display_url'), url=entity.get('expanded_url', user['legacy']['url']),
                            tcourl=user['legacy']['url'], indices=tuple(entity['indices']))
        label = None
        if (labelO := user['affiliates_highlighted_label'].get('label')):
            label = self._user_label_to_user_label(labelO)
        return User(
            username=user['legacy']['screen_name'],
            id=int(user['rest_id']),
            displayname=user['legacy']['name'],
            rawDescription=rawDescription,
            renderedDescription=renderedDescription,
            descriptionLinks=[TextLink(
                text=x.get('display_url'),
                url=x['expanded_url'],
                tcourl=x['url'],
                indices=tuple(x['indices']),
            ) for x in user['legacy']['entities']['description']['urls']],
            verified=user['legacy']['verified'],
            created=email.utils.parsedate_to_datetime(user['legacy']['created_at']),
            followersCount=user['legacy']['followers_count'],
            friendsCount=user['legacy']['friends_count'],
            statusesCount=user['legacy']['statuses_count'],
            favouritesCount=user['legacy']['favourites_count'],
            listedCount=user['legacy']['listed_count'],
            mediaCount=user['legacy']['media_count'],
            location=user['legacy']['location'],
            protected=user['legacy']['protected'],
            link=link,
            profileImageUrl=user['legacy']['profile_image_url_https'],
            profileBannerUrl=user['legacy'].get('profile_banner_url'),
            label=label,
        )

    def get_items(self):
        if self._isUserId:
            # Resolve user ID to username
            self._user = self.entity.username
            self._isUserId = False
            self._query = f'from:{self._user}'
        yield from super().get_items()

    @staticmethod
    def is_valid_username(s):
        return 1 <= len(s) <= 20 and s.strip(string.ascii_letters + string.digits + '_') == ''

    @classmethod
    def _cli_setup_parser(cls, subparser):
        def user(s):
            if cls.is_valid_username(s) or s.isdigit():
                return s
            raise ValueError('Invalid username or ID')

        subparser.add_argument('--user-id', dest='isUserId', action='store_true', default=False,
                               help='Use user ID instead of username')
        subparser.add_argument('user', type=user, help='A Twitter username (without @)')

    @classmethod
    def _cli_from_args(cls, args):
        return cls._cli_construct(args, user=int(args.user) if args.isUserId else args.user)


class TwitterProfileScraper(TwitterUserScraper):
    name = 'twitter-profile'

    def get_items(self):
        if not self._isUserId:
            userId = self.entity.id
        else:
            userId = self._user
        paginationVariables = {
            'userId': userId,
            'count': 100,
            'cursor': None,
            'includePromotedContent': True,
            'withCommunity': True,
            'withSuperFollowsUserFields': True,
            'withDownvotePerspective': False,
            'withReactionsMetadata': False,
            'withReactionsPerspective': False,
            'withSuperFollowsTweetFields': True,
            'withVoice': True,
            'withV2Timeline': False,
        }
        variables = paginationVariables.copy()
        del variables['cursor']
        features = {
            'responsive_web_twitter_blue_verified_badge_is_enabled': True,
            'verified_phone_label_enabled': False,
            'responsive_web_graphql_timeline_navigation_enabled': True,
            'view_counts_public_visibility_enabled': True,
            'view_counts_everywhere_api_enabled': True,
            'longform_notetweets_consumption_enabled': False,
            'tweetypie_unmention_optimization_enabled': True,
            'responsive_web_uc_gql_enabled': True,
            'vibe_api_enabled': True,
            'responsive_web_edit_tweet_api_enabled': True,
            'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
            'standardized_nudges_misinfo': True,
            'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': False,
            'interactive_text_enabled': True,
            'responsive_web_text_conversations_enabled': False,
            'responsive_web_enhance_cards_enabled': False,
        }

        params = {'variables': variables, 'features': features}
        paginationParams = {'variables': paginationVariables, 'features': features}

        gotPinned = False
        for obj in self._iter_api_data('https://twitter.com/i/api/graphql/W3HCLclD2VauuL6RcQm9MA/UserTweetsAndReplies',
                                       _TwitterAPIType.GRAPHQL, params, paginationParams):
            instructions = obj['data']['user']['result']['timeline']['timeline']['instructions']
            if not gotPinned:
                for instruction in instructions:
                    if instruction['type'] == 'TimelinePinEntry':
                        gotPinned = True
                        yield self._graphql_timeline_tweet_item_result_to_tweet(
                            instruction['entry']['content']['itemContent']['tweet_results']['result'])
            yield from self._graphql_timeline_instructions_to_tweets(instructions)


class TwitterHashtagScraper(TwitterSearchScraper):
    name = 'twitter-hashtag'

    def __init__(self, hashtag, **kwargs):
        super().__init__(f'#{hashtag}', **kwargs)
        self._hashtag = hashtag

    @classmethod
    def _cli_setup_parser(cls, subparser):
        subparser.add_argument('hashtag', type=umbraSnscrpe.base.nonempty_string('hashtag'),
                               help='A Twitter hashtag (without #)')

    @classmethod
    def _cli_from_args(cls, args):
        return cls._cli_construct(args, args.hashtag)


class TwitterTweetScraperMode(enum.Enum):
    SINGLE = 'single'
    SCROLL = 'scroll'
    RECURSE = 'recurse'

    @classmethod
    def _cli_from_args(cls, args):
        if args.scroll:
            return cls.SCROLL
        if args.recurse:
            return cls.RECURSE
        return cls.SINGLE


class TwitterTweetScraper(_TwitterAPIScraper):
    name = 'twitter-tweet'

    def __init__(self, tweetId, *, mode=TwitterTweetScraperMode.SINGLE, **kwargs):
        self._tweetId = tweetId
        self._mode = mode
        super().__init__(f'https://twitter.com/i/web/status/{self._tweetId}', **kwargs)

    def get_items(self):
        paginationVariables = {
            'focalTweetId': str(self._tweetId),
            'cursor': None,
            'referrer': 'tweet',
            'with_rux_injections': False,
            'includePromotedContent': True,
            'withCommunity': True,
            'withQuickPromoteEligibilityTweetFields': True,
            'withBirdwatchNotes': True,
            'withSuperFollowsUserFields': True,
            'withDownvotePerspective': False,
            'withReactionsMetadata': False,
            'withReactionsPerspective': False,
            'withSuperFollowsTweetFields': True,
            'withVoice': True,
            'withV2Timeline': False,
        }
        variables = paginationVariables.copy()
        del variables['cursor'], variables['referrer']
        features = {
            'responsive_web_twitter_blue_verified_badge_is_enabled': True,
            'verified_phone_label_enabled': False,
            'responsive_web_graphql_timeline_navigation_enabled': True,
            'view_counts_public_visibility_enabled': True,
            'view_counts_everywhere_api_enabled': True,
            'longform_notetweets_consumption_enabled': False,
            'tweetypie_unmention_optimization_enabled': True,
            'responsive_web_uc_gql_enabled': True,
            'vibe_api_enabled': True,
            'responsive_web_edit_tweet_api_enabled': True,
            'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
            'standardized_nudges_misinfo': True,
            'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': False,
            'interactive_text_enabled': True,
            'responsive_web_text_conversations_enabled': False,
            'responsive_web_enhance_cards_enabled': False,
        }

        params = {'variables': variables, 'features': features}
        paginationParams = {'variables': paginationVariables, 'features': features}
        url = 'https://twitter.com/i/api/graphql/HQ_gjq7zDNvSiJOCSkwUEw/TweetDetail'
        if self._mode is TwitterTweetScraperMode.SINGLE:
            obj = self._get_api_data(url, _TwitterAPIType.GRAPHQL, params=params)
            if not obj['data']:
                return
            for instruction in obj['data']['threaded_conversation_with_injections']['instructions']:
                if instruction['type'] != 'TimelineAddEntries':
                    continue
                for entry in instruction['entries']:
                    if entry['entryId'] == f'tweet-{self._tweetId}' and entry['content'][
                        'entryType'] == 'TimelineTimelineItem' and entry['content']['itemContent'][
                        'itemType'] == 'TimelineTweet':
                        yield self._graphql_timeline_tweet_item_result_to_tweet(
                            entry['content']['itemContent']['tweet_results']['result'])
                        break
        elif self._mode is TwitterTweetScraperMode.SCROLL:
            for obj in self._iter_api_data(url, _TwitterAPIType.GRAPHQL, params, paginationParams,
                                           direction=_ScrollDirection.BOTH):
                if not obj['data']:
                    continue
                yield from self._graphql_timeline_instructions_to_tweets(
                    obj['data']['threaded_conversation_with_injections']['instructions'],
                    includeConversationThreads=True)
        elif self._mode is TwitterTweetScraperMode.RECURSE:
            seenTweets = set()
            queue = collections.deque()
            queue.append(self._tweetId)
            while queue:
                tweetId = queue.popleft()
                thisPagParams = copy.deepcopy(paginationVariables)
                thisPagParams['variables']['focalTweetId'] = str(tweetId)
                thisParams = copy.deepcopy(thisPagParams)
                del thisPagParams['variables']['cursor'], thisPagParams['variables']['referrer']
                for obj in self._iter_api_data(url, _TwitterAPIType.GRAPHQL, thisParams, thisPagParams,
                                               direction=_ScrollDirection.BOTH):
                    if not obj['data']:
                        continue
                    for tweet in self._graphql_timeline_instructions_to_tweets(
                            obj['data']['threaded_conversation_with_injections']['instructions'],
                            includeConversationThreads=True):
                        if tweet.id not in seenTweets:
                            yield tweet
                            seenTweets.add(tweet.id)
                            if tweet.id != self._tweetId:  # Already queued at the beginning
                                queue.append(tweet.id)

    @classmethod
    def _cli_setup_parser(cls, subparser):
        group = subparser.add_mutually_exclusive_group(required=False)
        group.add_argument('--scroll', action='store_true', default=False, help='Enable scrolling in both directions')
        group.add_argument('--recurse', '--recursive', action='store_true', default=False,
                           help='Enable recursion through all tweets encountered (warning: slow, potentially memory-intensive!)')
        subparser.add_argument('tweetId', type=int, help='A tweet ID')

    @classmethod
    def _cli_from_args(cls, args):
        return cls._cli_construct(args, args.tweetId, mode=TwitterTweetScraperMode._cli_from_args(args))


class TwitterListPostsScraper(TwitterSearchScraper):
    name = 'twitter-list-posts'

    def __init__(self, listName, **kwargs):
        super().__init__(f'list:{listName}', **kwargs)
        self._listName = listName

    @classmethod
    def _cli_setup_parser(cls, subparser):
        subparser.add_argument('list', type=umbraSnscrpe.base.nonempty_string('list'),
                               help='A Twitter list ID or a string of the form "username/listname" (replace spaces with dashes)')

    @classmethod
    def _cli_from_args(cls, args):
        return cls._cli_construct(args, args.list)


class TwitterTrendsScraper(_TwitterAPIScraper):
    name = 'twitter-trends'

    def __init__(self, **kwargs):
        super().__init__('https://twitter.com/i/trends', **kwargs)

    def get_items(self):
        params = {
            'include_profile_interstitial_type': '1',
            'include_blocking': '1',
            'include_blocked_by': '1',
            'include_followed_by': '1',
            'include_want_retweets': '1',
            'include_mute_edge': '1',
            'include_can_dm': '1',
            'include_can_media_tag': '1',
            'include_ext_has_nft_avatar': '1',
            'skip_status': '1',
            'cards_platform': 'Web-12',
            'include_cards': '1',
            'include_ext_alt_text': 'true',
            'include_quote_count': 'true',
            'include_reply_count': '1',
            'tweet_mode': 'extended',
            'include_entities': 'true',
            'include_user_entities': 'true',
            'include_ext_media_color': 'true',
            'include_ext_media_availability': 'true',
            'include_ext_sensitive_media_warning': 'true',
            'include_ext_trusted_friends_metadata': 'true',
            'send_error_codes': 'true',
            'simple_quoted_tweet': 'true',
            'count': '20',
            'candidate_source': 'trends',
            'include_page_configuration': 'false',
            'entity_tokens': 'false',
            'ext': 'mediaStats,highlightedLabel,hasNftAvatar,voiceInfo,enrichments,superFollowMetadata,unmentionInfo',
        }
        obj = self._get_api_data('https://twitter.com/i/api/2/guide.json', _TwitterAPIType.V2, params)
        for instruction in obj['timeline']['instructions']:
            if not 'addEntries' in instruction:
                continue
            for entry in instruction['addEntries']['entries']:
                if entry['entryId'] != 'trends':
                    continue
                for item in entry['content']['timelineModule']['items']:
                    trend = item['item']['content']['trend']
                    yield Trend(name=trend['name'], metaDescription=trend['trendMetadata'].get('metaDescription'),
                                domainContext=trend['trendMetadata']['domainContext'])



__all__ = ['Toot', 'Boost', 'Attachment', 'Poll', 'PollOption', 'User', 'CustomEmoji', 'MastodonProfileScraper', 'MastodonTootScraperMode', 'MastodonTootScraper']


import bs4
import dataclasses
import datetime
import enum
import json
import logging
import umbraSnscrpe.base
import time
import typing
import urllib.parse


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Toot(umbraSnscrpe.base.Item):
	url: str
	id: str
	user: 'User'
	date: datetime.datetime
	text: str
	spoilerText: typing.Optional[str] = None
	attachments: typing.Optional[typing.List['Attachment']] = None
	links: typing.Optional[typing.List[str]] = None
	mentionedUsers: typing.Optional[typing.List['User']] = None
	hashtags: typing.Optional[typing.List[str]] = None
	poll: typing.Optional['Poll'] = None

	def __str__(self):
		return self.url


@dataclasses.dataclass
class Boost(umbraSnscrpe.base.Item):
	user: 'User'
	toot: Toot

	def __str__(self):
		# Boosts don't have their own URLs
		return str(self.toot)


@dataclasses.dataclass
class Attachment:
	url: str
	name: str


@dataclasses.dataclass
class Poll:
	id: str
	expirationDate: datetime.datetime
	multiple: bool
	options: typing.List['PollOption']
	votesCount: int
	votersCount: typing.Optional[int] = None # Available since version 3.0.0 (commit 3babf846)


@dataclasses.dataclass
class PollOption:
	title: str
	votesCount: int


@dataclasses.dataclass
class User(umbraSnscrpe.base.Entity):
	account: str # @username@domain.invalid
	displayName: typing.Optional[str] = None
	displayNameWithCustomEmojis: typing.Optional[typing.List[typing.Union[str, 'CustomEmoji']]] = None
	avatarUrl: typing.Optional[str] = None
	_url: typing.Optional[str] = None

	@property
	def url(self):
		if self._url:
			return self._url
		return f'https://{"/@".join(reversed(self.account[1:].split("@")))}'

	def __str__(self):
		return self.url


@dataclasses.dataclass
class CustomEmoji:
	shortName: str
	url: str
	staticUrl: str


class _MastodonCommonScraper(umbraSnscrpe.base.Scraper):
	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self._headers = {'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0', 'Accept-Language': 'en-US,en;q=0.5'}
		self._lastRequest = 0

	def _rate_limited_get(self, *args, **kwargs):
		if (diff := time.time() - self._lastRequest) < 3:
			time.sleep(3 - diff)
		self._lastRequest = time.time()
		return self._get(*args, **kwargs)

	def _entries_to_items(self, entries, url):
		for entry in entries:
			if entry.find('a', class_ = 'load-more'):
				continue

			tootKwargs = {}

			info = entry.find('div', class_ = 'status__info')
			if not info: # Before 2.5.0 (commit bb71538b)
				info = entry.find('div', class_ = 'status__header')
			if not info: # Detailed status (i.e. toot page rather than timeline)?
				info = entry.find('div', class_ = 'detailed-status__meta')
			link = info.find('a', class_ = 'status__relative-time')
			if not link: # Detailed status?
				link = info.find('a', class_ = 'detailed-status__datetime')
			tootKwargs['url'] = link['href']
			tootKwargs['id'] = tootKwargs['url'].rsplit('/', 1)[1]
			tootKwargs['date'] = datetime.datetime.strptime(info.find('data', class_ = 'dt-published')['value'], '%Y-%m-%dT%H:%M:%S+00:00').replace(tzinfo = datetime.timezone.utc)

			userKwargs = {}
			userLink = info.find('a', class_ = 'status__display-name')
			if not userLink: # Detailed status?
				userLink = entry.find('a', class_ = 'detailed-status__display-name')
			userNameSpan = userLink.find('span', class_ = 'display-name')
			userKwargs['account'] = userNameSpan.find('span').text.strip()
			if userKwargs['account'].count('@') == 1: # Ancient versions don't include the instance for posts from accounts on the instance itself
				userKwargs['account'] = self._url_to_account(userLink['href'])
			userKwargs['_url'] = urllib.parse.urljoin(url, userLink['href'])
			userKwargs['displayName'], userKwargs['displayNameWithCustomEmojis'] = self._display_name(userNameSpan.find('strong'), url)
			userKwargs['avatarUrl'] = urllib.parse.urljoin(url, userLink.find('img', class_ = 'u-photo')['src'])
			tootKwargs['user'] = User(**userKwargs)

			content = entry.find('div', class_ = 'status__content')
			if not content.find(class_ = 'status__content__spoiler-link'):
				tootKwargs['text'] = '\n\n'.join(p.text for p in content.find_all('p'))
			else:
				tootKwargs['text'] = content.find('span', class_ = 'p-summary').text
				tootKwargs['spoilerText'] = '\n\n'.join(p.text for p in content.find('div', class_ = 'e-content').find_all('p'))

			if (attachmentsDiv := entry.find('div', class_ = 'attachment-list')):
				attachments = []
				for a in attachmentsDiv.find_all('a'):
					attachments.append(Attachment(url = urllib.parse.urljoin(url, a['href']), name = a.text.strip()))
				tootKwargs['attachments'] = attachments
			elif (mediaGalleryDiv := entry.find('div', attrs = {'data-component': 'MediaGallery'})): # Before 2.7.0 (https://github.com/mastodon/mastodon/issues/6714)
				o = json.loads(mediaGalleryDiv['data-props'])
				attachments = []
				for medium in o['media']:
					attachments.append(Attachment(url = urllib.parse.urljoin(url, medium['url']), name = medium['url'].rsplit('/', 1)[-1].strip()))
				tootKwargs['attachments'] = attachments
			elif (attachmentsDiv := entry.find('div', class_ = 'status__attachments')): # Before 2.3.0 (commit 2bbf987a)
				attachments = []
				for a in attachmentsDiv.find_all('a'):
					attachments.append(Attachment(url = urllib.parse.urljoin(url, a['href']), name = a['href'].rsplit('/', 1)[1]))
				tootKwargs['attachments'] = attachments

			links = []
			mentionedUsers = []
			hashtags = []
			for a in content.find_all('a'):
				cls = a.get('class', [])
				if 'mention' in cls and 'u-url' in cls:
					mentionUrl = urllib.parse.urljoin(url, a['href'])
					mentionedUsers.append(User(account = self._url_to_account(mentionUrl), _url = mentionUrl))
				elif 'mention' in cls and 'hashtag' in cls:
					hashtags.append(a.text.strip())
				else:
					links.append(urllib.parse.urljoin(url, a['href']))
			if links:
				tootKwargs['links'] = links
			if mentionedUsers:
				tootKwargs['mentionedUsers'] = mentionedUsers
			if hashtags:
				tootKwargs['hashtags'] = hashtags

			if (pollDiv := entry.find('div', attrs = {'data-component': 'Poll'})):
				o = json.loads(pollDiv['data-props'])
				pollKwargs = {}
				pollKwargs['id'] = o['poll']['id']
				pollKwargs['expirationDate'] = datetime.datetime.strptime(o['poll']['expires_at'], '%Y-%m-%dT%H:%M:%S.%fZ').replace(tzinfo = datetime.timezone.utc)
				pollKwargs['multiple'] = o['poll']['multiple']
				pollKwargs['options'] = [PollOption(title = op['title'], votesCount = op['votes_count']) for op in o['poll']['options']]
				pollKwargs['votesCount'] = o['poll']['votes_count']
				if 'voters_count' in o['poll']: # 3.0.0 (commit 3babf846)
					pollKwargs['votersCount'] = o['poll']['voters_count']
				tootKwargs['poll'] = Poll(**pollKwargs)

			toot = Toot(**tootKwargs)

			# Boosts
			prepend = entry.find('div', class_ = 'status__prepend')
			if not prepend: # Before 2.5.0 (commit bb71538b)
				prepend = entry.find('div', class_ = 'pre-header')
			if prepend and prepend.find('i', class_ = 'fa-retweet'): # Is a boost
				userKwargs = {}
				userLink = prepend.find('a', class_ = 'status__display-name')
				# The user is always on this instance since that's the only place where boosts are shown, hence there is no explicit account span. Reconstruct from URL.
				userUrl = urllib.parse.urljoin(url, userLink['href'])
				assert userUrl.count('/') == 3 and userUrl.count('/@') == 1
				userKwargs['account'] = '@'.join(reversed(userUrl.split('/')[2:]))
				userKwargs['displayName'], userKwargs['displayNameWithCustomEmojis'] = self._display_name(userLink.find('strong'), url)
				toot = Boost(user = User(**userKwargs), toot = toot)

			yield toot

	def _display_name(self, strong, url):
		outPlain = []
		outFull = []
		hasCustomEmoji = False
		for child in strong.children:
			if isinstance(child, bs4.element.NavigableString):
				outPlain.append(str(child))
				outFull.append(str(child))
			elif child.name == 'img' and 'custom-emoji' in child.get('class', []):
				hasCustomEmoji = True
				outPlain.append(child['alt'])
				outFull.append(CustomEmoji(shortName = child['alt'], url = urllib.parse.urljoin(url, child['data-original']), staticUrl = urllib.parse.urljoin(url, child['data-static'])))
			elif child.name == 'img' and 'emojione' in child.get('class', []):
				# Version 2.0.0 (which first added custom emojis) to 2.9.4: no data-* attributes, only gets one of the URLs with no (easy, reliable) way of knowing which it is.
				hasCustomEmoji = True
				outPlain.append(child['alt'])
				outFull.append(CustomEmoji(shortName = child['alt'], url = urllib.parse.urljoin(url, child['src'])))
			else:
				_logger.warning(f'Unexpected display name child: {child!r}')
		return ''.join(outPlain), outFull if hasCustomEmoji else None

	@staticmethod
	def _url_to_account(url):
		if url.count('/') == 3 and url.count('/@') == 1:
			return '@'.join(reversed(url.split('/')[2:]))
		if url.count('/') == 4 and '/users/' in url: # E.g. Pleroma, also supported by Mastodon
			return '@' + '@'.join(reversed(url.split('/')[2::2]))
		if url.count('/') == 4 and '/accounts/' in url: # E.g. Peertube
			return '@' + '@'.join(reversed(url.split('/')[2::2]))
		if url.count('/') == 4 and '/profile/' in url: # E.g. Friendica
			return '@' + '@'.join(reversed(url.split('/')[2::2]))
		raise ValueError('Unrecognised account URL format')


class MastodonProfileScraper(_MastodonCommonScraper):
	name = 'mastodon-profile'

	def __init__(self, account, **kwargs):
		super().__init__(**kwargs)
		if account.startswith('@') and account.count('@') == 2:
			account, domain = account[1:].split('@')
			url = f'https://{domain}/@{account}'
		else:
			url = account
		self._url = url

	def get_items(self):
		initial = True
		while True:
			if initial:
				r = self._rate_limited_get(f'{self._url}/with_replies', headers = self._headers)
				if r.status_code not in (200, 404):
					raise umbraSnscrpe.base.ScraperException(f'Got status code {r.status_code}')
				if r.status_code == 404: # Possibly an old instance where with_replies doesn't exist, try without that.
					r = self._rate_limited_get(self._url, headers = self._headers)
					if r.status_code not in (200, 404):
						raise umbraSnscrpe.base.ScraperException(f'Got status code {r.status_code}')
					if r.status_code == 404:
						_logger.warning('Account does not exist')
						return
					_logger.warning('Old Mastodon instance, cannot retrieve reply toots')
				initial = False
			else:
				r = self._rate_limited_get(url, headers = self._headers)
				if r.status_code != 200:
					raise umbraSnscrpe.base.ScraperException(f'Got status code {r.status_code}')
			soup = bs4.BeautifulSoup(r.text, 'lxml')

			yield from self._entries_to_items(soup.find('div', class_ = 'activity-stream').find_all('div', class_ = 'entry'), r.url)

			nextA = soup.find('a', class_ = 'load-more', href = lambda x: '?max_id=' in x or '&max_id=' in x)
			if not nextA: # Before 2.5.0 (commit bb71538b)
				paginationDiv = soup.find('div', class_ = 'pagination')
				if paginationDiv:
					nextA = paginationDiv.find('a', class_ = 'next')
			if not nextA: # End of pagination
				break
			url = urllib.parse.urljoin(r.url, nextA['href'])

	@classmethod
	def _cli_setup_parser(cls, subparser):
		subparser.add_argument('account', type = umbraSnscrpe.base.nonempty_string('account'), help ='A Mastodon account. This can be either a URL to the profile page or a string of the form @account@instance.example.org')

	@classmethod
	def _cli_from_args(cls, args):
		return cls._cli_construct(args, args.account)


class MastodonTootScraperMode(enum.Enum):
	SINGLE = 'single'
	THREAD = 'thread'

	@classmethod
	def _cli_from_args(cls, args):
		if args.thread:
			return cls.THREAD
		return cls.SINGLE


class MastodonTootScraper(_MastodonCommonScraper):
	name = 'mastodon-toot'

	def __init__(self, url, *, mode = MastodonTootScraperMode.SINGLE, **kwargs):
		super().__init__(**kwargs)
		self._url = url
		self._mode = mode

	def get_items(self):
		r = self._rate_limited_get(self._url, headers = self._headers)
		if r.status_code == 404:
			_logger.warning('Toot does not exist')
			return
		if r.status_code != 200:
			raise umbraSnscrpe.base.ScraperException(f'Got status code {r.status_code}')
		soup = bs4.BeautifulSoup(r.text, 'lxml')
		if self._mode is MastodonTootScraperMode.SINGLE:
			status = soup.find('div', class_ = 'detailed-status')
			entry = status.parent
			yield from self._entries_to_items([entry], r.url)
		elif self._mode is MastodonTootScraperMode.THREAD:
			yield from self._entries_to_items(soup.find('div', class_ = 'activity-stream').find_all('div', class_ = 'entry'), r.url)

	@classmethod
	def _cli_setup_parser(cls, subparser):
		subparser.add_argument('--thread', action = 'store_true', help = 'Collect thread around the toot referenced by the URL')
		subparser.add_argument('url', type = umbraSnscrpe.base.nonempty_string('url'), help ='A URL for a toot')

	@classmethod
	def _cli_from_args(cls, args):
		return cls._cli_construct(args, args.url, mode = MastodonTootScraperMode._cli_from_args(args))



import abc
import copy
import dataclasses
import datetime
import functools
import json
import logging
import requests
import time
import warnings


logger = logging.getLogger(__name__)


class _DeprecatedProperty:
	def __init__(self, name, repl, replStr):
		self.name = name
		self.repl = repl
		self.replStr = replStr

	def __get__(self, obj, objType):
		if obj is None: # if the access is through the class using _DeprecatedProperty rather than an instance of the class:
			return self
		warnings.warn(f'{self.name} is deprecated, use {self.replStr} instead', FutureWarning, stacklevel = 2)
		return self.repl(obj)


def _json_serialise_datetime(obj):
	'''A JSON serialiser that converts datetime.datetime and datetime.date objects to ISO-8601 strings.'''

	if isinstance(obj, (datetime.datetime, datetime.date)):
		return obj.isoformat()
	raise TypeError(f'Object of type {type(obj)} is not JSON serializable')


def _json_dataclass_to_dict(obj):
	if isinstance(obj, _JSONDataclass) or dataclasses.is_dataclass(obj):
		out = {}
		out['_type'] = f'{type(obj).__module__}.{type(obj).__name__}'
		for field in dataclasses.fields(obj):
			assert field.name != '_type'
			if field.name.startswith('_'):
				continue
			out[field.name] = _json_dataclass_to_dict(getattr(obj, field.name))
		# Add in (non-deprecated) properties
		for k in dir(obj):
			if isinstance(getattr(type(obj), k, None), property):
				assert k != '_type'
				if k.startswith('_'):
					continue
				out[k] = _json_dataclass_to_dict(getattr(obj, k))
		return out
	elif isinstance(obj, (tuple, list)):
		return type(obj)(_json_dataclass_to_dict(x) for x in obj)
	elif isinstance(obj, dict):
		return {_json_dataclass_to_dict(k): _json_dataclass_to_dict(v) for k, v in obj.items()}
	elif isinstance(obj, set):
		return {_json_dataclass_to_dict(v) for v in obj}
	else:
		return copy.deepcopy(obj)


@dataclasses.dataclass
class _JSONDataclass:
	'''A base class for dataclasses for conversion to JSON'''

	def json(self):
		'''Convert the object to a JSON string'''

		out = _json_dataclass_to_dict(self)
		for key, value in list(out.items()): # Modifying the dict below, so make a copy first
			if isinstance(value, IntWithGranularity):
				out[key] = int(value)
				assert f'{key}.granularity' not in out, f'Granularity collision on {key}.granularity'
				out[f'{key}.granularity'] = value.granularity
		return json.dumps(out, default = _json_serialise_datetime)


@dataclasses.dataclass
class Item(_JSONDataclass):
	'''An abstract base class for an item returned by the scraper's get_items generator.

	An item can really be anything. The string representation should be useful for the CLI output (e.g. a direct URL for the item).
	'''

	@abc.abstractmethod
	def __str__(self):
		pass


@dataclasses.dataclass
class Entity(_JSONDataclass):
	'''An abstract base class for an entity returned by the scraper's entity property.

	An entity is typically the account of a person or organisation. The string representation should be the preferred direct URL to the entity's page on the network.
	'''

	@abc.abstractmethod
	def __str__(self):
		pass


class IntWithGranularity(int):
	'''A number with an associated granularity

	For example, an IntWithGranularity(42000, 1000) represents a number on the order of 42000 with two significant digits, i.e. something counted with a granularity of 1000.
	'''

	def __new__(cls, value, granularity, *args, **kwargs):
		obj = super().__new__(cls, value, *args, **kwargs)
		obj.granularity = granularity
		return obj

	def __reduce__(self):
		return (IntWithGranularity, (int(self), self.granularity))


class URLItem(Item):
	'''A generic item which only holds a URL string.'''

	def __init__(self, url):
		self._url = url

	@property
	def url(self):
		return self._url

	def __str__(self):
		return self._url


class ScraperException(Exception):
	pass


class Scraper:
	'''An abstract base class for a scraper.'''

	name = None

	def __init__(self, *, retries = 3, proxies = None):
		self._retries = retries
		self._proxies = proxies
		self._session = requests.Session()

	@abc.abstractmethod
	def get_items(self):
		'''Iterator yielding Items.'''

		pass

	def _get_entity(self):
		'''Get the entity behind the scraper, if any.

		This is the method implemented by subclasses for doing the actual retrieval/entity object creation. For accessing the scraper's entity, use the entity property.
		'''

		return None

	@functools.cached_property
	def entity(self):
		return self._get_entity()

	def _request(self, method, url, params = None, data = None, headers = None, timeout = 10, responseOkCallback = None, allowRedirects = True, proxies = None):
		proxies = proxies or self._proxies or {}
		errors = []
		for attempt in range(self._retries + 1):
			# The request is newly prepared on each retry because of potential cookie updates.
			req = self._session.prepare_request(requests.Request(method, url, params = params, data = data, headers = headers))
			environmentSettings = self._session.merge_environment_settings(req.url, proxies, None, None, None)
			logger.info(f'Retrieving {req.url}')
			logger.debug(f'... with headers: {headers!r}')
			if data:
				logger.debug(f'... with data: {data!r}')
			if environmentSettings:
				logger.debug(f'... with environmentSettings: {environmentSettings!r}')
			try:
				r = self._session.send(req, allow_redirects = allowRedirects, timeout = timeout, **environmentSettings)
			except requests.exceptions.RequestException as exc:
				if attempt < self._retries:
					retrying = ', retrying'
					level = logging.INFO
				else:
					retrying = ''
					level = logging.ERROR
				logger.log(level, f'Error retrieving {req.url}: {exc!r}{retrying}')
				errors.append(repr(exc))
			else:
				redirected = f' (redirected to {r.url})' if r.history else ''
				logger.info(f'Retrieved {req.url}{redirected}: {r.status_code}')
				if r.history:
					for i, redirect in enumerate(r.history):
						logger.debug(f'... request {i}: {redirect.request.url}: {r.status_code} (Location: {r.headers.get("Location")})')
				if responseOkCallback is not None:
					success, msg = responseOkCallback(r)
					errors.append(msg)
				else:
					success, msg = (True, None)
				msg = f': {msg}' if msg else ''

				if success:
					logger.debug(f'{req.url} retrieved successfully{msg}')
					return r
				else:
					if attempt < self._retries:
						retrying = ', retrying'
						level = logging.INFO
					else:
						retrying = ''
						level = logging.ERROR
					logger.log(level, f'Error retrieving {req.url}{msg}{retrying}')
			if attempt < self._retries:
				sleepTime = 1.0 * 2**attempt # exponential backoff: sleep 1 second after first attempt, 2 after second, 4 after third, etc.
				logger.info(f'Waiting {sleepTime:.0f} seconds')
				time.sleep(sleepTime)
		else:
			msg = f'{self._retries + 1} requests to {req.url} failed, giving up.'
			logger.fatal(msg)
			logger.fatal(f'Errors: {", ".join(errors)}')
			raise ScraperException(msg)
		raise RuntimeError('Reached unreachable code')

	def _get(self, *args, **kwargs):
		return self._request('GET', *args, **kwargs)

	def _post(self, *args, **kwargs):
		return self._request('POST', *args, **kwargs)

	@classmethod
	def _cli_setup_parser(cls, subparser):
		pass

	@classmethod
	def _cli_from_args(cls, args):
		return cls._cli_construct(args)

	@classmethod
	def _cli_construct(cls, argparseArgs, *args, **kwargs):
		return cls(*args, **kwargs, retries = argparseArgs.retries)


def nonempty_string(name):
	def f(s):
		s = s.strip()
		if s:
			return s
		raise ValueError('must not be an empty string')
	f.__name__ = name
	return f



__all__ = ['InstagramPost', 'User', 'InstagramUserScraper', 'InstagramHashtagScraper', 'InstagramLocationScraper']


import dataclasses
import datetime
import hashlib
import json
import logging
import re
import umbraSnscrpe.base
import typing


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class InstagramPost(umbraSnscrpe.base.Item):
	url: str
	date: datetime.datetime
	content: typing.Optional[str]
	thumbnailUrl: str
	displayUrl: str
	username: typing.Optional[str]
	likes: int
	comments: int
	commentsDisabled: bool
	isVideo: bool

	def __str__(self):
		return self.url


@dataclasses.dataclass
class User(umbraSnscrpe.base.Entity):
	username: str
	name: typing.Optional[str]
	followers: umbraSnscrpe.base.IntWithGranularity
	following: umbraSnscrpe.base.IntWithGranularity
	posts: umbraSnscrpe.base.IntWithGranularity

	followersGranularity = umbraSnscrpe.base._DeprecatedProperty('followersGranularity', lambda self: self.followers.granularity, 'followers.granularity')
	followingGranularity = umbraSnscrpe.base._DeprecatedProperty('followingGranularity', lambda self: self.following.granularity, 'following.granularity')
	postsGranularity = umbraSnscrpe.base._DeprecatedProperty('postsGranularity', lambda self: self.posts.granularity, 'posts.granularity')

	def __str__(self):
		return f'https://www.instagram.com/{self.username}/'


class _InstagramCommonScraper(umbraSnscrpe.base.Scraper):
	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self._headers = {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'}
		self._initialPage = None

	def _response_to_items(self, response):
		for node in response[self._responseContainer][self._edgeXToMedia]['edges']:
			code = node['node']['shortcode']
			username = node['node']['owner']['username'] if 'username' in node['node']['owner'] else None
			url = f'https://www.instagram.com/p/{code}/'
			yield InstagramPost(
			  url = url,
			  date = datetime.datetime.fromtimestamp(node['node']['taken_at_timestamp'], datetime.timezone.utc),
			  content = node['node']['edge_media_to_caption']['edges'][0]['node']['text'] if len(node['node']['edge_media_to_caption']['edges']) else None,
			  thumbnailUrl = node['node']['thumbnail_src'],
			  displayUrl = node['node']['display_url'],
			  username = username,
			  likes = node['node']['edge_media_preview_like']['count'],
			  comments = node['node']['edge_media_to_comment']['count'],
			  commentsDisabled = node['node']['comments_disabled'],
			  isVideo = node['node']['is_video'],
			 )

	def _initial_page(self):
		if self._initialPage is None:
			_logger.info('Retrieving initial data')
			r = self._get(self._initialUrl, headers = self._headers, responseOkCallback = self._check_initial_page_callback)
			if r.status_code not in (200, 404):
				raise umbraSnscrpe.base.ScraperException(f'Got status code {r.status_code}')
			elif r.url.startswith('https://www.instagram.com/accounts/login/'):
				raise umbraSnscrpe.base.ScraperException('Redirected to login page')
			self._initialPage = r
		return self._initialPage

	def _check_initial_page_callback(self, r):
		if r.status_code != 200:
			return True, None
		jsonData = r.text.split('<script type="text/javascript">window._sharedData = ')[1].split(';</script>')[0] # May throw an IndexError if Instagram changes something again; we just let that bubble.
		try:
			obj = json.loads(jsonData)
		except json.JSONDecodeError:
			return False, 'invalid JSON'
		r._snscrape_json_obj = obj
		return True, None

	def _check_json_callback(self, r):
		if r.status_code != 200:
			return False, f'status code {r.status_code}'
		if r.url.startswith('https://www.instagram.com/accounts/login/'):
			raise umbraSnscrpe.base.ScraperException('Redirected to login page')
		try:
			obj = json.loads(r.text)
		except json.JSONDecodeError as e:
			return False, f'invalid JSON ({e!r})'
		r._snscrape_json_obj = obj
		return True, None

	def get_items(self):
		r = self._initial_page()
		if r.status_code == 404:
			_logger.warning('Page does not exist')
			return
		response = r._snscrape_json_obj
		rhxGis = response['rhx_gis'] if 'rhx_gis' in response else ''
		if response['entry_data'][self._pageName][0]['graphql'][self._responseContainer][self._edgeXToMedia]['count'] == 0:
			_logger.info('Page has no posts')
			return
		if not response['entry_data'][self._pageName][0]['graphql'][self._responseContainer][self._edgeXToMedia]['edges']:
			_logger.warning('Private account')
			return
		pageID = response['entry_data'][self._pageName][0]['graphql'][self._responseContainer][self._pageIDKey]
		yield from self._response_to_items(response['entry_data'][self._pageName][0]['graphql'])
		if not response['entry_data'][self._pageName][0]['graphql'][self._responseContainer][self._edgeXToMedia]['page_info']['has_next_page']:
			return
		endCursor = response['entry_data'][self._pageName][0]['graphql'][self._responseContainer][self._edgeXToMedia]['page_info']['end_cursor']

		headers = self._headers.copy()
		while True:
			_logger.info(f'Retrieving endCursor = {endCursor!r}')
			variables = self._variablesFormat.format(**locals())
			headers['X-Requested-With'] = 'XMLHttpRequest'
			headers['X-Instagram-GIS'] = hashlib.md5(f'{rhxGis}:{variables}'.encode('utf-8')).hexdigest()
			r = self._get(f'https://www.instagram.com/graphql/query/?query_hash={self._queryHash}&variables={variables}', headers = headers, responseOkCallback = self._check_json_callback)

			if r.status_code != 200:
				raise umbraSnscrpe.base.ScraperException(f'Got status code {r.status_code}')

			response = r._snscrape_json_obj
			if not response['data'][self._responseContainer][self._edgeXToMedia]['edges']:
				return
			yield from self._response_to_items(response['data'])
			if not response['data'][self._responseContainer][self._edgeXToMedia]['page_info']['has_next_page']:
				return
			endCursor = response['data'][self._responseContainer][self._edgeXToMedia]['page_info']['end_cursor']


class InstagramUserScraper(_InstagramCommonScraper):
	name = 'instagram-user'

	def __init__(self, username, **kwargs):
		super().__init__(**kwargs)
		self._initialUrl = f'https://www.instagram.com/{username}/'
		self._pageName = 'ProfilePage'
		self._responseContainer = 'user'
		self._edgeXToMedia = 'edge_owner_to_timeline_media'
		self._pageIDKey = 'id'
		self._queryHash = 'f2405b236d85e8296cf30347c9f08c2a'
		self._variablesFormat = '{{"id":"{pageID}","first":50,"after":"{endCursor}"}}'

	def _get_entity(self):
		r = self._initial_page()
		if r.status_code != 200:
			return
		if '<meta property="og:description" content="' not in r.text:
			return
		ogDescriptionContentPos = r.text.index('<meta property="og:description" content="') + len('<meta property="og:description" content="')
		ogDescription = r.text[ogDescriptionContentPos : r.text.index('"', ogDescriptionContentPos)]

		numPattern = r'\d+(?:\.\d+)?m|\d+(?:\.\d+)?k|\d+,\d+|\d+'
		ogDescriptionPattern = re.compile('^(' + numPattern + ') Followers, (' + numPattern + ') Following, (' + numPattern + r') Posts - See Instagram photos and videos from (?:(.*?) \(@([a-z0-9_.]+)\)|@([a-z0-9_-]+))$')
		m = ogDescriptionPattern.match(ogDescription)
		assert m, 'unexpected og:description format'

		def parse_num(s):
			if s.endswith('m'):
				return int(float(s[:-1].replace(',', '')) * 1e6), 10 ** (6 if '.' not in s else 6 - len(s[:-1].replace(',', '').split('.')[1]))
			elif s.endswith('k'):
				return int(float(s[:-1].replace(',', '')) * 1000), 10 ** (3 if '.' not in s else 3 - len(s[:-1].replace(',', '').split('.')[1]))
			else:
				return int(s.replace(',', '')), 1

		followers = umbraSnscrpe.base.IntWithGranularity(*parse_num(m.group(1)))
		following = umbraSnscrpe.base.IntWithGranularity(*parse_num(m.group(2)))
		posts = umbraSnscrpe.base.IntWithGranularity(*parse_num(m.group(3)))
		return User(
			username = m.group(5) or m.group(6),
			name = m.group(4) or None,
			followers = followers,
			following = following,
			posts = posts,
		  )

	@classmethod
	def _cli_setup_parser(cls, subparser):
		subparser.add_argument('username', type = umbraSnscrpe.base.nonempty_string('username'), help ='An Instagram username (no leading @)')

	@classmethod
	def _cli_from_args(cls, args):
		return cls._cli_construct(args, args.username)


class InstagramHashtagScraper(_InstagramCommonScraper):
	name = 'instagram-hashtag'

	def __init__(self, hashtag, **kwargs):
		super().__init__(**kwargs)
		self._initialUrl = f'https://www.instagram.com/explore/tags/{hashtag}/'
		self._pageName = 'TagPage'
		self._responseContainer = 'hashtag'
		self._edgeXToMedia = 'edge_hashtag_to_media'
		self._pageIDKey = 'name'
		self._queryHash = 'f92f56d47dc7a55b606908374b43a314'
		self._variablesFormat = '{{"tag_name":"{pageID}","first":50,"after":"{endCursor}"}}'

	@classmethod
	def _cli_setup_parser(cls, subparser):
		subparser.add_argument('hashtag', type = umbraSnscrpe.base.nonempty_string('hashtag'), help ='An Instagram hashtag (no leading #)')

	@classmethod
	def _cli_from_args(cls, args):
		return cls._cli_construct(args, args.hashtag)


class InstagramLocationScraper(_InstagramCommonScraper):
	name = 'instagram-location'

	def __init__(self, locationId, **kwargs):
		super().__init__(**kwargs)
		self._initialUrl = f'https://www.instagram.com/explore/locations/{locationId}/'
		self._pageName = 'LocationsPage'
		self._responseContainer = 'location'
		self._edgeXToMedia = 'edge_location_to_media'
		self._pageIDKey = 'id'
		self._queryHash = '1b84447a4d8b6d6d0426fefb34514485'
		self._variablesFormat = '{{"id":"{pageID}","first":50,"after":"{endCursor}"}}'

	@classmethod
	def _cli_setup_parser(cls, subparser):
		subparser.add_argument('locationid', help = 'An Instagram location ID', type = int)

	@classmethod
	def _cli_from_args(cls, args):
		return cls._cli_construct(args, args.locationid)



from setuptools import setup, find_packages
import codecs
import os


VERSION = '0.1.0'
DESCRIPTION = ''
LONG_DESCRIPTION = ''

# Setting up
setup(
    name="umbraSnscrpe",
    version=VERSION,
    author="umbra",
    author_email="",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['filelock', 'pyautogui', 'pyaudio'],
    keywords=['lxml', 'requests[socks]', 'beautifulsoup4', 'pytz; python_version < "3.9.0"', 'filelock'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ]
)


__all__ = ['Submission', 'Comment', 'RedditUserScraper', 'RedditSubredditScraper', 'RedditSearchScraper', 'RedditSubmissionScraper']


import dataclasses
import datetime
import logging
import re
import umbraSnscrpe.base
import umbraSnscrpe.version
import string
import time
import typing


_logger = logging.getLogger(__name__)


# Most of these fields should never be None, but due to broken data, they sometimes are anyway...

@dataclasses.dataclass
class Submission(umbraSnscrpe.base.Item):
	author: typing.Optional[str] # E.g. submission hf7k6
	date: datetime.datetime
	id: str
	link: typing.Optional[str]
	selftext: typing.Optional[str]
	subreddit: typing.Optional[str] # E.g. submission 617p51
	title: str
	url: str

	created = umbraSnscrpe.base._DeprecatedProperty('created', lambda self: self.date, 'date')

	def __str__(self):
		return self.url


@dataclasses.dataclass
class Comment(umbraSnscrpe.base.Item):
	author: typing.Optional[str]
	body: str
	date: datetime.datetime
	id: str
	parentId: typing.Optional[str]
	subreddit: typing.Optional[str]
	url: str

	created = umbraSnscrpe.base._DeprecatedProperty('created', lambda self: self.date, 'date')

	def __str__(self):
		return self.url


def _cmp_id(id1, id2):
	'''Compare two Reddit IDs. Returns -1 if id1 is less than id2, 0 if they are equal, and 1 if id1 is greater than id2.

	id1 and id2 may have prefixes like t1_, but if included, they must be present on both and equal.'''

	if id1.startswith('t') and '_' in id1:
		prefix, id1 = id1.split('_', 1)
		if not id2.startswith(f'{prefix}_'):
			raise ValueError('id2 must have the same prefix as id1')
		_, id2 = id2.split('_', 1)
	if id1.strip(string.ascii_lowercase + string.digits) != '':
		raise ValueError('invalid characters in id1')
	if id2.strip(string.ascii_lowercase + string.digits) != '':
		raise ValueError('invalid characters in id2')
	if len(id1) < len(id2):
		return -1
	if len(id1) > len(id2):
		return 1
	if id1 < id2:
		return -1
	if id1 > id2:
		return 1
	return 0


class _RedditPushshiftScraper(umbraSnscrpe.base.Scraper):
	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self._headers = {'User-Agent': f'snscrape/{umbraSnscrpe.version.__version__}'}

	def _handle_rate_limiting(self, r):
		if r.status_code == 429:
			_logger.info('Got 429 response, sleeping')
			time.sleep(10)
			return False, 'rate-limited'
		if r.status_code != 200:
			return False, 'non-200 status code'
		return True, None

	def _get_api(self, url, params = None):
		r = self._get(url, params = params, headers = self._headers, responseOkCallback = self._handle_rate_limiting)
		if r.status_code != 200:
			raise umbraSnscrpe.base.ScraperException(f'Got status code {r.status_code}')
		return r.json()

	def _api_obj_to_item(self, d):
		cls = Submission if 'title' in d else Comment

		# Pushshift doesn't always return a permalink; sometimes, there's a permalink_url instead, and sometimes there's nothing at all
		permalink = d.get('permalink')
		if permalink is None:
			# E.g. comment dovj2v7
			permalink = d.get('permalink_url')
			if permalink is None:
				if 'link_id' in d and d['link_id'].startswith('t3_'): # E.g. comment doraazf
					if 'subreddit' in d:
						permalink = f'/r/{d["subreddit"]}/comments/{d["link_id"][3:]}/_/{d["id"]}/'
					else: # E.g. submission 617p51 but can likely happen for comments as well
						permalink = f'/comments/{d["link_id"][3:]}/_/{d["id"]}/'
				else:
					_logger.warning('Unable to find or construct permalink')
					permalink = '/'

		kwargs = {
			'author': d.get('author'),
			'date': datetime.datetime.fromtimestamp(d['created_utc'], datetime.timezone.utc),
			'url': f'https://old.reddit.com{permalink}',
			'subreddit': d.get('subreddit'),
		}
		if cls is Submission:
			kwargs['selftext'] = d.get('selftext') or None
			kwargs['link'] = (d['url'] if not d['url'].startswith('/') else f'https://old.reddit.com{d["url"]}') if not kwargs['selftext'] else None
			if kwargs['link'] == kwargs['url'] or kwargs['url'].replace('//old.reddit.com/', '//www.reddit.com/') == kwargs['link']:
				kwargs['link'] = None
			kwargs['title'] = d['title']
			kwargs['id'] = f't3_{d["id"]}'
		else:
			kwargs['body'] = d['body']
			kwargs['parentId'] = d.get('parent_id')
			kwargs['id'] = f't1_{d["id"]}'

		return cls(**kwargs)

	def _iter_api(self, url, params = None):
		'''Iterate through the Pushshift API using the 'until' parameter and yield the items.'''
		lowestIdSeen = None
		if params is None:
			params = {}
		while True:
			obj = self._get_api(url, params = params)
			if not obj['data'] or (lowestIdSeen is not None and all(_cmp_id(d['id'], lowestIdSeen) >= 0 for d in obj['data'])): # end of pagination
				break
			for d in obj['data']:
				if lowestIdSeen is None or _cmp_id(d['id'], lowestIdSeen) == -1:
					yield self._api_obj_to_item(d)
					lowestIdSeen = d['id']
			params['until'] = obj["data"][-1]["created_utc"] + 1


class _RedditPushshiftSearchScraper(_RedditPushshiftScraper):
	def __init__(self, name, *, submissions = True, comments = True, before = None, after = None, **kwargs):
		super().__init__(**kwargs)
		self._name = name
		self._submissions = submissions
		self._comments = comments
		self._before = before
		self._after = after

		if not type(self)._validationFunc(self._name):
			raise ValueError(f'invalid {type(self).name.split("-", 1)[1]} name')
		if not self._submissions and not self._comments:
			raise ValueError('At least one of submissions and comments must be True')

	def _iter_api_submissions_and_comments(self, params: dict):
		# Retrieve both submissions and comments, interleave the results to get a reverse-chronological order
		params['limit'] = '1000'
		if self._before is not None:
			params['until'] = self._before
		if self._after is not None:
			params['since'] = self._after

		if self._submissions:
			submissionsIter = self._iter_api('https://api.pushshift.io/reddit/search/submission', params.copy()) # Pass copies to prevent the two iterators from messing each other up by using the same dict
		else:
			submissionsIter = iter(())
		if self._comments:
			commentsIter = self._iter_api('https://api.pushshift.io/reddit/search/comment', params.copy())
		else:
			commentsIter = iter(())

		try:
			tipSubmission = next(submissionsIter)
		except StopIteration:
			# There are no submissions, just yield comments and return
			yield from commentsIter
			return
		try:
			tipComment = next(commentsIter)
		except StopIteration:
			# There are no comments, just yield submissions and return
			yield tipSubmission
			yield from submissionsIter
			return

		while True:
			# Return newer first; if both have the same creation datetime, return the comment first
			if tipSubmission.date > tipComment.date:
				yield tipSubmission
				try:
					tipSubmission = next(submissionsIter)
				except StopIteration:
					# Reached the end of submissions, just yield the remaining comments and stop
					yield tipComment
					yield from commentsIter
					break
			else:
				yield tipComment
				try:
					tipComment = next(commentsIter)
				except StopIteration:
					yield tipSubmission
					yield from submissionsIter
					break

	def get_items(self):
		yield from self._iter_api_submissions_and_comments({type(self)._apiField: self._name})

	@classmethod
	def _cli_setup_parser(cls, subparser):
		subparser.add_argument('--no-submissions', dest = 'noSubmissions', action = 'store_true', default = False, help = 'Don\'t list submissions')
		subparser.add_argument('--no-comments', dest = 'noComments', action = 'store_true', default = False, help = 'Don\'t list comments')
		subparser.add_argument('--before', metavar = 'TIMESTAMP', type = int, help = 'Fetch results before a Unix timestamp')
		subparser.add_argument('--after', metavar = 'TIMESTAMP', type = int, help = 'Fetch results after a Unix timestamp')
		name = cls.name.split('-', 1)[1]
		subparser.add_argument(name, type = umbraSnscrpe.base.nonempty_string(name))

	@classmethod
	def _cli_from_args(cls, args):
		name = cls.name.split('-', 1)[1]
		return cls._cli_construct(args, getattr(args, name), submissions = not args.noSubmissions, comments = not args.noComments, before = args.before, after = args.after)


class RedditUserScraper(_RedditPushshiftSearchScraper):
	name = 'reddit-user'
	_validationFunc = lambda x: re.match('^[A-Za-z0-9_-]{3,20}$', x)
	_apiField = 'author'


class RedditSubredditScraper(_RedditPushshiftSearchScraper):
	name = 'reddit-subreddit'
	_validationFunc = lambda x: re.match('^[A-Za-z0-9][A-Za-z0-9_]{2,20}$', x)
	_apiField = 'subreddit'


class RedditSearchScraper(_RedditPushshiftSearchScraper):
	name = 'reddit-search'
	_validationFunc = lambda x: True
	_apiField = 'q'


class RedditSubmissionScraper(_RedditPushshiftScraper):
	name = 'reddit-submission'

	def __init__(self, submissionId, **kwargs):
		if (submissionId[3:] if submissionId.startswith('t3_') else submissionId).strip(string.ascii_lowercase + string.digits) != '':
			raise ValueError('invalid submissionId')
		super().__init__(**kwargs)
		self._submissionId = submissionId

	def get_items(self):
		obj = self._get_api(f'https://api.pushshift.io/reddit/search/submission?ids={self._submissionId}')
		if not obj['data']:
			return
		if len(obj['data']) != 1:
			raise umbraSnscrpe.base.ScraperException(f'Got {len(obj["data"])} results instead of 1')
		yield self._api_obj_to_item(obj['data'][0])

		# Upstream bug: link_id must be provided in decimal https://old.reddit.com/r/pushshift/comments/zkggt0/update_on_colo_switchover_bug_fixes_reindexing/
		yield from self._iter_api('https://api.pushshift.io/reddit/search/comment', {'link_id': int(self._submissionId, 36), 'limit': 1000})

	@classmethod
	def _cli_setup_parser(cls, subparser):
		subparser.add_argument('submissionId', type = umbraSnscrpe.base.nonempty_string('submissionId'))

	@classmethod
	def _cli_from_args(cls, args):
		return cls._cli_construct(args, args.submissionId)
